import os
from typing import List, Dict
import admix
import pandas as pd
import dapgen
import numpy as np
import dask.array as da
from tqdm import tqdm
from ._utils import log_params


def _append_to_file(path: str, df: pd.DataFrame):
    """
    Append a new data frame to an existing .pvar file. The index of the new data frame
    is assumed to be exactly the same as the index of the existing data frame.

    Parameters
    ----------
    pvar_path : str
        Path to the .pvar file.
    df : pd.DataFrame
        Data frame to append.

    """
    # if path does not exist, directly write to it
    if not os.path.exists(path):
        df.to_csv(path, sep="\t", float_format="%.8g")
    else:
        df_snp_info = pd.read_csv(path, sep="\t", index_col=0)
        assert df_snp_info.index.equals(
            df.index
        ), f"SNPs in {path} and the provided data frame do not match"
        # no overlap of columns
        overlap_cols = set(df_snp_info.columns) & set(df.columns)
        assert len(overlap_cols) == 0, f"Overlap of columns: {overlap_cols}"
        df_snp_info = pd.merge(df_snp_info, df, left_index=True, right_index=True)
        df_snp_info.to_csv(path, sep="\t", float_format="%.8g")


def grm(
    plink_file: str,
    out_prefix: str,
    subpopu: str = None,
    std_method: str = "std",
    snp_chunk_size: int = 256,
    snp_list: str = None,
) -> None:
    """
    Calculate the GRM for a given PLINK file

    Parameters
    ----------
    plink_file : str
        Path to the pfile
    out_prefix : str
        Prefix of the output files
    subpopu : str
        Path to the subpopulation file
    std_method : str
        Method to standardize the GRM. Currently supported:
        "std" (standardize to have mean 0 and variance 1),
        "allele" (standardize to have mean 0 but no scaling)
    snp_chunk_size : int, optional
        Number of SNPs to read at a time, by default 256
        This can be tuned to reduce memory usage
    snp_list : str, optional
        Path to a file containing a list of SNPs to use. Each line should be a SNP ID.
        Only SNPs in the list will be used for the analysis. By default None

    Returns
    -------
    GRM files: {out_prefix}.[grm.bin | grm.id | grm.n] will be generated
    Weight file: {out_prefix}.weight.tsv will be generated
    """
    assert std_method in ["std", "allele"], f"Unknown std_method: {std_method}"
    log_params("grm", locals())

    geno, df_snp, df_indiv = dapgen.read_plink(plink_file, snp_chunk=snp_chunk_size)
    n_indiv = len(df_indiv)

    if subpopu is not None:
        df_subpopu = pd.read_csv(
            "data/subpopu.txt", delim_whitespace=True, header=None, dtype=str
        )
        df_subpopu.columns = ["FID", "IID", "POPU"]
        df_indiv = pd.merge(df_indiv, df_subpopu, on=["FID", "IID"])
        assert (
            len(df_indiv) == n_indiv
        ), "Individuals in the subpopulation file do not match the PLINK file"

    # filter for SNPs
    snp_subset = np.ones(len(df_snp)).astype(bool)
    if snp_list is not None:
        with open(snp_list, "r") as f:
            filter_snp_list = [line.strip() for line in f]
        n_filter_snp = len(filter_snp_list)
        snp_subset = snp_subset & df_snp.index.isin(filter_snp_list)
        if sum(snp_subset) < n_filter_snp:
            admix.logger.warning(
                f"{n_filter_snp - sum(snp_subset)} SNPs in {snp_list} are not in the dataset"
            )

    admix.logger.info(f"{sum(snp_subset)} SNPs are used for GRM calculation")

    # subset
    df_snp = df_snp.loc[snp_subset, :]
    geno = geno[snp_subset, :]

    # calculate GRM
    if subpopu is None:
        grm = admix.data.grm(geno, subpopu=None, std_method=std_method)
    else:
        grm = admix.data.grm(
            geno, subpopu=df_indiv["POPU"].values, std_method=std_method
        )

    df_id = pd.DataFrame({"0": df_indiv.FID.values, "1": df_indiv.IID.values})

    # write GRM
    admix.tools.gcta.write_grm(
        out_prefix,
        K=grm,
        df_id=df_id,
        n_snps=np.repeat(len(df_snp), len(df_id)),
    )


def append_snp_info(
    pfile: str,
    out: str = None,
    info: List[str] = ["LANC_FREQ", "FREQ"],
):
    """
    Append information to .pvar file. Currently, 3 statistics are supported:
    (1) ancestry-specific frequency (2) ancestry-specific haplotype count (3) total
    allele frequency is supported. Please raise an issue if you need other statistics.

    Parameters
    ----------
    pfile : str
        Path to the .pvar file.
    out : str
        Path to the output file. If specified, the output file will be WRITTEN to
        this path. Otherwise, the output file will be appended to the <pfile>.snp_info
        file.
    info : List[str]
        List of information to append. Currently supported:

        * "LANC_FREQ": ancestry-specific allele frequency of each SNP. For example, for a two-way admixture population, `LANC_FREQ1` indicates the frequency of alternate allele in the first ancestry. `LANC_NHAPLO1` will also be added, indicating the number of haplotypes in the first ancestry.
        * "FREQ": allele frequency of each SNP regardless of ancestry.

    Examples
    --------
    .. code-block:: bash
        
        # toy-admix.snp_info will be created containing LANC_FREQ[1-n_anc], LANC_NHAPLO[1-n_anc], FREQ
        admix append-snp-info \\
            --pfile toy-admix \\
            --out toy-admix.snp_info
         
    """
    log_params("append-snp-info", locals())

    dset = admix.io.read_dataset(pfile)

    df_info: pd.DataFrame = []
    if "LANC_FREQ" in info:
        af = dset.af_per_anc()
        nhaplo = dset.nhaplo_per_anc()

        df_af = pd.DataFrame(
            af,
            columns=[f"LANC_FREQ{i + 1}" for i in range(af.shape[1])],
            index=dset.snp.index,
        )
        df_info.append(df_af)

        df_nhaplo = pd.DataFrame(
            nhaplo,
            columns=[f"LANC_NHAPLO{i + 1}" for i in range(nhaplo.shape[1])],
            index=dset.snp.index,
        )
        df_info.append(df_nhaplo)

    if "FREQ" in info:
        df_freq = dapgen.freq(pfile + ".pgen", memory=8)
        assert np.all(df_freq.ID.values == dset.snp.index)
        df_freq = pd.DataFrame(
            df_freq["ALT_FREQS"].values,
            columns=["FREQ"],
            index=dset.snp.index,
        )
        df_info.append(df_freq)

    df_info = pd.concat(df_info, axis=1)

    if out is None:
        out = pfile + ".snp_info"

    _append_to_file(out, df_info)


def da_sum(mat: da.Array):
    """Summation over a dask array to get individual-level sum

    Parameters
    ----------
    mat : da.Array
        dask array to sum over
    """
    chunks = mat.chunks[0]
    indices = np.insert(np.cumsum(chunks), 0, 0)
    mat_sum = np.zeros(mat.shape[1])
    for i in range(len(indices) - 1):
        start, stop = indices[i], indices[i + 1]
        mat_sum += mat[start:stop, :].sum(axis=0).compute()
    return mat_sum


def calc_pgs(
    plink_path: str,
    weights_path: str,
    out: str,
    weight_col: str = "WEIGHT",
):
    """Calculate PGS from a weight file and a PLINK file.

    Parameters
    ----------
    plink_path : str
        Path to plink files. Format examples:
          * /path/to/chr21.pgen
          * /path/to/genotype/directory
          * /path/to/file_list.txt # file_list.txt contains rows of file names
    weights_path : str
        path to PGS weights, containing CHROM, SNP, REF, ALT, WEIGHT columns
    out : str
        prefix of the output files.
    weight_col : str, optional
        column in 'weights_path' representing the weight, by default "WEIGHT"
    """
    log_params("calc-pgs", locals())
    df_weights = pd.read_csv(weights_path, sep="\t")
    pgen_files = dapgen.parse_plink_path(plink_path)
    if not isinstance(pgen_files, list):
        pgen_files = [pgen_files]
    plink_prefix_list = [pgen.rsplit(".", 1)[0] for pgen in pgen_files]

    df_snp_info: Dict = {"PLINK_SNP": [], "WEIGHTS_SNP": [], "WEIGHTS": []}
    all_partial_pgs = 0.0
    all_lanc = 0.0
    indiv_list = None
    all_n_snp = 0
    for plink_prefix in plink_prefix_list:
        dset = admix.io.read_dataset(plink_prefix)
        assert dset.n_anc == 2, "Only 2-way admixture is currently supported"
        if indiv_list is None:
            indiv_list = dset.indiv.index
        else:
            assert indiv_list.equals(
                dset.indiv.index
            ), f"Indiv list in {plink_prefix} does not match with previous ones."
        dset_idx, wgt_idx, flip = dapgen.align_snp(
            df1=dset.snp[["CHROM", "POS", "REF", "ALT"]], df2=df_weights
        )
        dset_subset = dset[dset_idx]
        admix.logger.info(f"# matched SNPs={len(wgt_idx)} for dset={plink_prefix}.")
        tmp_df_weights = df_weights[[weight_col]].loc[wgt_idx, :] * flip.reshape(-1, 1)
        tmp_df_weights.index = dset_idx

        partial_pgs = admix.data.calc_pgs(
            dset=dset_subset, df_weights=tmp_df_weights, method="partial"
        )

        df_snp_info["PLINK_SNP"].extend(dset_idx)
        df_snp_info["WEIGHTS_SNP"].extend(wgt_idx)
        df_snp_info["WEIGHTS"].extend(df_weights.loc[wgt_idx, weight_col].values * flip)
        all_partial_pgs += partial_pgs
        all_lanc += da_sum(dset_subset.lanc.sum(axis=2))
        all_n_snp += dset_subset.n_snp

    pd.DataFrame(df_snp_info).to_csv(out + ".snp_info.tsv", sep="\t", index=False)
    pd.DataFrame(
        {
            "indiv": dset.indiv.index,
            "PGS1": all_partial_pgs[:, 0],
            "PGS2": all_partial_pgs[:, 1],
            "PROP1": 1 - all_lanc / (all_n_snp * 2),
            "PROP2": all_lanc / (all_n_snp * 2),
        }
    ).to_csv(out + ".pgs.tsv", sep="\t", index=False)
    admix.logger.info(f"SNP information saved to {out}.snp_info.tsv")
    admix.logger.info(f"PGS saved to {out}.pgs.tsv")


def calc_partial_pgs(
    plink_path: str,
    weights_path: str,
    out: str,
    ref_plink_path: str = None,
    ref_pops: List[str] = None,
    weight_col: str = "WEIGHT",
    ref_pop_col: str = "Population",
    dset_build: str = None,
    weights_build: str = None,
):
    """Calculate PGS from a weight file and a PLINK file.

    Parameters
    ----------
    plink_path : str
        Path to plink files. Some examples are:
          * :code:`/path/to/chr21.pgen`
          * :code:`/path/to/genotype/directory`
          * :code:`/path/to/file_list.txt` containing rows of file names
    weights_path : str
        path to PGS weights, containing :code:`CHROM`, :code:`SNP`, :code:`REF`,
        :code:`ALT`, :code:`WEIGHT` columns
    out : str
        prefix of the output files. :code:`{out}.sample_pgs.tsv` and
        :code:`{out}.ref_pgs.tsv` will be written to disk.
    ref_plink_path : str
        path to reference plink files. The :code:`ref_plink_path` should be a single
        plink2 file.
    ref_pops: list of str
        list of populations in reference plink files. For example, "CEU,YRI"
    weight_col : str, optional
        column in 'weights_path' representing the weight, by default "WEIGHT"
    dset_build: str, optional
        genome build transform for the dataset ("hg19->hg38" or "hg38->hg19"),
        by default None
    weights_build: str, optional
        genome build transform for the weights ("hg19->hg38" or "hg38->hg19"),
        by default None
    """
    log_params("calc-partial-pgs", locals())

    CHECK_COLS = ["CHROM", "POS", "REF", "ALT"]
    # read input data & basic checks
    pgen_files = dapgen.parse_plink_path(plink_path)
    if not isinstance(pgen_files, list):
        pgen_files = [pgen_files]
    plink_prefix_list = [pgen.rsplit(".", 1)[0] for pgen in pgen_files]

    # scoring weights
    df_weights = pd.read_csv(weights_path, sep="\t")
    df_weights = df_weights[CHECK_COLS + [weight_col]].copy()
    if weights_build is not None:
        df_weights["POS"] = admix.tools.liftover.run(
            df_weights[["CHROM", "POS"]], chain=weights_build
        )
        df_weights = df_weights[df_weights.POS != -1].copy()

    assert (ref_plink_path is None) == (
        ref_pops is None
    ), "Either both or none of ref_plink_path and ref_pops should be provided."

    # reference data
    CALC_REF = ref_plink_path is not None

    if CALC_REF:
        assert ref_plink_path.endswith(".pgen") or ref_plink_path.endswith(
            ".bed"
        ), "Reference plink file should be a single .pgen or .bed file."
        # remove .pgen or .bed from the file name
        ref_plink_path = ref_plink_path.rsplit(".", 1)[0]
        admix.logger.info(f"Reading reference plink file from {ref_plink_path}")
        dset_ref = admix.io.read_dataset(ref_plink_path)
        ref_pop_indiv: Dict = {
            pop: dset_ref.indiv.index[dset_ref.indiv[ref_pop_col] == pop].values
            for pop in ref_pops
        }

        admix.logger.info(
            f"Reading reference data with "
            + ", ".join(
                [f"{pop} #indiv={len(ref_pop_indiv[pop])}" for pop in ref_pop_indiv]
            )
        )

    total_sample_pgs: pd.DataFrame = 0
    if CALC_REF:
        total_ref_pgs: Dict[str, pd.DataFrame] = {pop: 0 for pop in ref_pops}
    # iterate through each plink file
    for i, plink_prefix in enumerate(plink_prefix_list):
        admix.logger.info(
            f"Scoring for dataset {i + 1}/{len(plink_prefix_list)}: {plink_prefix}"
        )
        _dset = admix.io.read_dataset(plink_prefix)
        if dset_build is not None:
            _dset.snp["POS"] = admix.tools.liftover.run(
                _dset.snp[["CHROM", "POS"]], chain=dset_build
            )
        _n_total_snp = _dset.n_snp
        assert _dset.n_anc == 2, "Only 2-way admixture is currently supported"
        # align sample dset and weights
        idx1, idx2, flip = dapgen.align_snp(
            df1=_dset.snp[CHECK_COLS], df2=df_weights[CHECK_COLS]
        )
        _dset = _dset[idx1]
        _df_weights = df_weights.loc[idx2, :].copy()
        _df_weights.index = idx1

        if CALC_REF:
            # align sample dset and reference dset
            idx1, idx2, flip = dapgen.align_snp(
                df1=_dset.snp[CHECK_COLS],
                df2=dset_ref.snp[CHECK_COLS],
            )
            _dset = _dset[idx1]
            _df_weights = _df_weights.loc[idx1, :]
            # original code: _dset_ref = dset_ref[idx2]
            # directly call admix.Dataset for potential unsorted scenarios:
            _dset_ref = admix.Dataset(
                dset_ref=dset_ref,
                snp_idx=dset_ref.snp.index.get_indexer(idx2),
                indiv_idx=slice(None),
                enforce_order=False,
            )
        admix.logger.info(f"matched #SNPs={_dset.n_snp}/{_n_total_snp}")

        if CALC_REF:
            sample_pgs, ref_pgs = admix.data.calc_partial_pgs(
                dset=_dset,
                df_weights=_df_weights,
                dset_ref=_dset_ref,
                ref_pop_indiv=[ref_pop_indiv[pop] for pop in ref_pops],
            )
            total_sample_pgs += sample_pgs
            for i, pop in enumerate(ref_pops):
                total_ref_pgs[pop] += ref_pgs[i]
        else:
            sample_pgs = admix.data.calc_partial_pgs(dset=_dset, df_weights=_df_weights)
            total_sample_pgs += sample_pgs

    total_sample_pgs.to_csv(out + ".sample_pgs.tsv", sep="\t")
    admix.logger.info(f"Sample PGS saved to {out}.sample_pgs.tsv")

    if CALC_REF:
        for pop in ref_pops:
            total_ref_pgs[pop].to_csv(out + f".ref_pgs_{pop}.tsv", sep="\t")
            admix.logger.info(f"Reference PGS saved to {out}.ref_pgs_{pop}.tsv")



import numpy as np
from numpy import (
    float32,
    tril_indices_from,
)
import pandas as pd
import dask.array as da
import admix
from typing import Union
import dapgen


def write_dataset(
    geno: da.Array,
    lanc: admix.data.Lanc,
    df_indiv: pd.DataFrame,
    df_snp: pd.DataFrame,
    out_prefix: str,
) -> None:

    n_snp, n_indiv = geno.shape[0:2]
    assert lanc.n_snp == n_snp
    assert lanc.n_indiv == n_indiv
    assert df_indiv.shape[0] == n_indiv
    assert df_snp.shape[0] == n_snp

    # write genotype
    dapgen.write_pgen(f"{out_prefix}.pgen", geno)

    # write local ancestry
    admix.io.write_lanc(f"{out_prefix}.lanc", lanc)

    # write df_indiv
    df_indiv.rename_axis("#IID").to_csv(f"{out_prefix}.psam", sep="\t")

    # write df_snp
    df_snp = df_snp.reset_index().rename(columns={"snp": "ID", "CHROM": "#CHROM"})

    FIXED_COLS = ["#CHROM", "POS", "ID", "REF", "ALT"]
    df_snp = df_snp[
        FIXED_COLS + [col for col in df_snp.columns if col not in FIXED_COLS]
    ]

    df_snp.to_csv(f"{out_prefix}.pvar", sep="\t", index=False)


def write_lanc(path: str, lanc: Union[np.ndarray, da.Array, admix.data.Lanc]) -> None:
    """
    Write local ancestry matrix to file.

    Parameters
    ----------
    path : str
        The path to the file to write.
    lanc : array_like, shape (n_snps, n_snps)
        The local ancestry matrix.
    """
    if isinstance(lanc, np.ndarray) or isinstance(lanc, da.Array):
        lanc = admix.data.Lanc(array=lanc)

    assert isinstance(
        lanc, admix.data.Lanc
    ), "lanc must be a numpy array or a dask array, or admix.data.Lanc"
    lanc.write(path)


def write_digit_mat(path, mat):
    """
    Read a matrix of integer with [0-9], and with no delimiter.

    Args
    ----

    """
    np.savetxt(path, mat, fmt="%d", delimiter="")


def write_gcta_grm(
    file_prefix: str, grm: np.ndarray, df_id: pd.DataFrame, n_snps: np.ndarray
) -> None:
    """
    Write a GCTA grm file.

    Parameters
    ----------
    file_prefix : str
        The prefix of the file to write.
    grm : array_like, shape (n_snps, n_snps)
        The GRM matrix.
    df_id : array_like, shape (n_indiv,)
        The ID of each individual.
    n_snps : np.ndarray, shape (n_indiv,)
        The number of SNPs for each individual.

    Returns
    -------
    None
    """

    bin_file = file_prefix + ".grm.bin"
    N_file = file_prefix + ".grm.N.bin"
    id_file = file_prefix + ".grm.id"

    # id
    df_id.to_csv(id_file, sep="\t", header=None, index=False)
    # bin
    grm[tril_indices_from(grm)].astype(float32).tofile(bin_file)
    # N
    n_snps.astype(float32).tofile(N_file)


# def write_gcta_pheno(
#     file_prefix: str, df_pheno: pd.DataFrame, df_id: pd.DataFrame
# ) -> None:
#     """
#     Write a GCTA phenotype file.

#     Parameters
#     ----------
#     file_prefix : str
#         The prefix of the file to write.
#     df_pheno : array_like, shape (n_indiv, n_pheno)
#         The phenotype matrix.
#     df_id : array_like, shape (n_indiv,)
#         The ID of each individual.

#     Returns
#     -------
#     None
#     """

#     pheno_file = file_prefix + ".pheno"
#     id_file = file_prefix + ".id"

#     # id
#     df_id.to_csv(id_file, sep="\t", header=None, index=False)
#     # pheno
#     df_pheno.to_csv(pheno_file, sep="\t", header=None, index=False)


import numpy as np
import pandas as pd
import tempfile
from os.path import join
import subprocess
from ..io import read_digit_mat, write_digit_mat


def lampld_wrapper(
    ref_hap_list,
    sample_hap,
    snp_pos,
    window_size=300,
    n_proto=6,
    bin_dir="/u/project/pasaniuc/kangchen/admix-tools/cache/LAMPLD-v1.3/bin/",
):
    tmp = tempfile.TemporaryDirectory()
    tmp_dir = tmp.name

    assert len(ref_hap_list) in [2, 3], "Currently only support 2-way / 3-way ancestry"
    for ref_i in range(len(ref_hap_list)):
        write_digit_mat(join(tmp_dir, f"ref{ref_i}.hap"), ref_hap_list[ref_i])
    write_digit_mat(join(tmp_dir, "sample.hap"), sample_hap)
    np.savetxt(join(tmp_dir, "pos.txt"), snp_pos, fmt="%s")

    if len(ref_hap_list) == 2:
        bin_path = join(bin_dir, "haplanc2way")
    elif len(ref_hap_list) == 3:
        bin_path = join(bin_dir, "haplanc")
    else:
        raise NotImplementedError
    cmd = " ".join(
        [
            bin_path,
            str(window_size),
            str(n_proto),
            join(tmp_dir, "pos.txt"),
            *[join(tmp_dir, f"ref{ref_i}.hap") for ref_i in range(len(ref_hap_list))],
            join(tmp_dir, "sample.hap"),
            join(tmp_dir, "out.txt"),
            str(1),
        ]
    )
    print(cmd)
    subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    infer_lanc = read_digit_mat(join(tmp_dir, "out.txt"))
    tmp.cleanup()
    return infer_lanc



import numpy as np
import dask.array as da
import pandas as pd
import xarray as xr
import warnings
from pandas.api.types import is_string_dtype
from typing import List, Tuple
import admix


def index_over_chunks(chunks: List[int]):
    """
    iterate over chunks of a dask array

    Parameters
    ----------
    chunks: List[int]
        Number of rows or columns in each chunk
    axis: int
        axis to iterate over 0 for row (SNP by convention), 1 for column (individual by convention)

    Returns
    -------
    generator
        generator of chunk indices so one can access the chunk with
        mat[start : stop, :] (axis=0) or mat[:, start : stop] (axis=1)
    """
    indices = np.insert(np.cumsum(chunks), 0, 0)
    for i in range(len(indices) - 1):
        start, stop = indices[i], indices[i + 1]
        yield start, stop


def match_prs_weights(
    dset: admix.Dataset, df_weight: pd.DataFrame, weight_cols: List[str]
) -> Tuple[admix.Dataset, pd.DataFrame]:
    # TODO: implement this with dapgen.align_snp
    """
    align the dataset and PRS weights with the following 3 steps:
    1. match SNPs in `dset` and `df_weight` by `CHROM` and `POS`
    2. Try match REF and ALT columns in `dset` and `df_weight`, either
        REF_dset = REF_weight, ALT_dset = ALT_weight, or
        REF_dset = ALT_weight, ALT_dset = REF_weight
    3. For mismatched SNPs,all columns in `weight_cols` of returned data frame
        will be reversed

    A shallow copy of `dset` will be returned.
    A copy of `df_weight` will be returned.
    Parameters
    ----------
    dset: a dataset with prs
    df_weight: a dataframe with prs weights
    weight_cols: list of columns in df_weight representing PRS weights
    """

    df_match_dset = (
        dset[["CHROM", "POS", "REF", "ALT"]]
        .to_dataframe()
        .reset_index()
        .rename(columns={"snp": "SNP"})
    )

    df_match_weight = df_weight[["SNP", "CHROM", "POS", "REF", "ALT"]].reset_index()

    df_merged = pd.merge(
        df_match_dset,
        df_match_weight,
        on=["CHROM", "POS"],
        suffixes=["_dset", "_weight"],
    )
    noflip_index = (df_merged["REF_dset"] == df_merged["REF_weight"]) & (
        df_merged["ALT_dset"] == df_merged["ALT_weight"]
    )
    flip_index = (df_merged["REF_dset"] == df_merged["ALT_weight"]) & (
        df_merged["ALT_dset"] == df_merged["REF_weight"]
    )

    rls_df_weight = df_weight.loc[
        df_weight.SNP.isin(df_merged["SNP_weight"][noflip_index | flip_index]),
        ["SNP", "CHROM", "POS", "REF", "ALT"] + weight_cols,
    ].copy()
    rls_df_weight.loc[
        rls_df_weight.SNP.isin(df_merged["SNP_weight"][flip_index]), weight_cols
    ] *= -1

    # the following step can be a bit slow sometimes, need to make sure both array
    # are like <U18 types to be fast
    rls_dset = dset.sel(
        snp=np.isin(
            dset.snp.values,
            df_merged["SNP_dset"][noflip_index | flip_index].values.astype(str),
        )
    )
    return rls_dset, rls_df_weight


def impute_lanc_old(dset: admix.Dataset, dset_ref: admix.Dataset):
    """
    Impute local ancestry using a reference dataset. The two data sets are assumed to
    have the same haplotype order, etc. Typically they are just a subset of each other.

    Using the following steps:
        1. basic checks are performed for the two data sets.
        2. `dset_ref`'s individuals is matched with `dset`, `dset`'s individuals
            must be a subset of `dset_ref`'s individuals.
        3. Imputation are performed

    Parameters
    ----------
    dset: a data set to be imputed with local ancestry
    dset_ref: a data set with local ancestry for reference

    Returns
    -------
    dset_imputed: a data set with imputed local ancestry
    """
    assert (
        len(set(dset.coords["CHROM"].data)) == 1
    ), "Data set to be imputed can only have one chromosome"

    # dset.indiv is a subset of dset_ref.indiv
    assert set(dset.indiv.values) <= set(dset_ref.indiv.values)
    dset_ref = dset_ref.sel(indiv=dset.indiv.values).sel(
        snp=(dset_ref.coords["CHROM"] == dset.coords["CHROM"][0])
    )

    # find relevant regions in reference dataset with local ancestry (hapmap3 SNPs here)
    # ref_start = np.argmin(np.abs(dset_ref["POS"].values - dset["POS"].values[0]))
    ref_start = np.where(dset_ref["POS"] < dset["POS"][0])[0][-1]
    # ref_stop = np.argmin(np.abs(dset_ref["POS"].values - dset["POS"].values[-1]))
    ref_stop = np.where(dset_ref["POS"] > dset["POS"][-1])[0][0]

    dset_ref_subset = dset_ref.isel(snp=slice(ref_start, ref_stop + 1))
    dset_ref_margin = dset_ref_subset.isel(snp=[0, -1])

    imputed_lanc = []
    for ploidy_i in range(2):
        df_ref_margin = pd.DataFrame(
            dset_ref_margin.lanc[:, ploidy_i].values.T,
            columns=dset_ref_margin.indiv.values,
            index=dset_ref_margin.snp.values,
        )
        df_imputed = pd.DataFrame(
            {
                "snp": dset.snp["snp"],
            }
        ).set_index("snp")
        df_imputed = pd.concat(
            [
                df_imputed,
                pd.DataFrame(columns=dset["indiv"].values, dtype=float),
            ]
        )
        # fill margin
        df_imputed = pd.concat(
            [df_ref_margin.iloc[[0], :], df_imputed, df_ref_margin.iloc[[-1], :]],
            axis=0,
        )
        df_imputed.index.name = "snp"
        # fill inside
        df_imputed.loc[
            dset_ref_subset.snp.values, dset_ref_subset.indiv.values
        ] = dset_ref_subset["lanc"][:, :, ploidy_i].values.T
        # interpolate
        df_imputed = (
            df_imputed.reset_index().interpolate(method="nearest").set_index("snp")
        )

        imputed_lanc.append(
            df_imputed.loc[dset["snp"].values, dset["indiv"].values]
            .values.astype(np.int8)
            .T
        )

    dset = dset.assign(
        lanc=(
            ("indiv", "snp", "ploidy"),
            da.from_array(np.dstack(imputed_lanc), chunks=-1),
        )
    )

    return dset


def impute_std(geno, mean=None, std=None):
    """
    impute the mean and then standardize
    geno: (num_indivs x num_snps) numpy array
    """
    if mean is None and std is None:
        mean = np.nanmean(geno, axis=0)
        nanidx = np.where(np.isnan(geno))
        geno[nanidx] = mean[nanidx[1]]
        std = np.std(geno, axis=0)
        std_geno = (geno - mean) / std
    else:
        nanidx = np.where(np.isnan(geno))
        geno[nanidx] = mean[nanidx[1]]
        std_geno = (geno - mean) / std
    return std_geno


def mean_std(geno, chunk_size=500):
    row_count = geno.row_count
    col_count = geno.col_count
    mean = np.zeros(col_count)
    std = np.zeros(col_count)

    for i in range(0, col_count, chunk_size):
        sub_geno = geno[:, i : i + chunk_size].read().val
        sub_mean = np.nanmean(sub_geno, axis=0)
        mean[i : i + chunk_size] = sub_mean
        nanidx = np.where(np.isnan(sub_geno))
        sub_geno[nanidx] = sub_mean[nanidx[1]]
        std[i : i + chunk_size] = np.std(sub_geno, axis=0)
    df = pd.DataFrame({"mean": mean, "std": std})
    return df


def cov(geno, mean_std, chunk_size=500):
    """
    geno: a Bed object which we are interested in estimating the LD
    chunk_size: number of SNPs to compute at a time
    """
    # first compute the mean and standard deviation for the given geno
    num_indv, num_snps = geno.shape
    cov = np.zeros([num_snps, num_snps])
    # compute mean and standard deviation
    for row_start in range(0, num_snps, chunk_size):
        for col_start in range(0, num_snps, chunk_size):
            # for each block
            row_stop = row_start + chunk_size
            col_stop = col_start + chunk_size
            if row_stop > num_snps:
                row_stop = num_snps
            if col_stop > num_snps:
                col_stop = num_snps

            std_row_geno = impute_std(
                geno[:, row_start:row_stop].read().val,
                mean_std["mean"][row_start:row_stop].values,
                mean_std["std"][row_start:row_stop].values,
            )
            std_col_geno = impute_std(
                geno[:, col_start:col_stop].read().val,
                mean_std["mean"][col_start:col_stop].values,
                mean_std["std"][col_start:col_stop].values,
            )

            cov[
                np.ix_(np.arange(row_start, row_stop), np.arange(col_start, col_stop))
            ] = np.dot(std_row_geno.T, std_col_geno) / (std_row_geno.shape[0])
    return cov


def make_dataset(
    geno, snp: pd.DataFrame, indiv: pd.DataFrame, meta: dict = None, lanc=None
):
    """Make up a dataset

    Parameters
    ----------
    geno : (#indiv, #snp, 2) array_like
        Genotype count matrix
    snp : pd.DataFrame
        Index is identifier to SNP
        each column corresponds to an attribute
    indiv : pd.DataFrame
        Index is identifier to individuals
        each column corresponds to an attribute
    meta : dict, optional
        Meta information about the dataset
    lanc : (#indiv, #snp, 2) array_like, optional
        Local ancestry
    """
    # fill `data_vars`
    data_vars: dict = {
        "geno": (("indiv", "snp", "ploidy"), geno),
    }
    if lanc is not None:
        data_vars["lanc"] = (("indiv", "snp", "ploidy"), lanc)

    coords: dict = {}
    # fill SNP information
    coords["snp"] = snp.index.values
    if not is_string_dtype(coords["snp"]):
        warnings.warn("Transforming snp index to str")
    coords["snp"] = coords["snp"].astype(str)

    for col in snp.columns:
        vals = snp[col].values
        if is_string_dtype(snp[col]):
            vals = snp[col].values.astype(str)

        coords[col] = ("snp", vals)

    # fill in individual information
    coords["indiv"] = indiv.index.values
    if not is_string_dtype(coords["indiv"]):
        warnings.warn("Transforming indiv index to str")
    coords["indiv"] = coords["indiv"].astype(str)

    for col in indiv.columns:
        vals = indiv[col].values
        if is_string_dtype(indiv[col]):
            vals = vals.astype(str)
        coords[col] = ("indiv", vals)

    dset = xr.Dataset(data_vars=data_vars, coords=coords, attrs=meta)
    return dset


def seperate_ld_blocks(anc, phgeno, legend, ld_blocks):
    assert len(legend) == anc.shape[1]
    assert len(legend) == phgeno.shape[1]

    rls_list = []
    for block_i, block in ld_blocks.iterrows():
        block_index = np.where(
            (block.START <= legend.position) & (legend.position < block.STOP)
        )[0]
        block_legend = legend.loc[block_index]
        block_anc = anc[:, block_index]
        block_phgeno = phgeno[:, block_index]
        rls_list.append((block_anc, block_phgeno, block_legend))
    return rls_list


def distance_to_line(
    p: np.ndarray, a: np.ndarray, b: np.ndarray, weight: np.ndarray = None
):
    """
    Calculate the distance of each row of a matrix to a reference line

    https://stackoverflow.com/questions/50727961/shortest-distance-between-a-point-and-a-line-in-3-d-space

    Parameters
    ----------
    p: np.ndarray
        matrix (n_indiv, n_var)
    a: np.ndarray
        ref point 1
    b: np.ndarray
        ref point 2
    weight: np.ndarray, optional
        weight associated to each dimension, for example, could be sqrt(eigenvalues)
        if not provided, equal weights for each coordinate will be used

    Returns
    -------
    dist: np.ndarray
        vector distance to the line
    t: np.ndarray
        projected length on the line, normalized according to the length of (b - a)
        t = 0 corresponds projection to a; t = 1 corresponds projection to b.
    n: np.ndarray
        normal vector from the projection point to the original point
        t*b + (1 - t)*a + n can be used to reconstruct the point position.
    """
    if weight is not None:
        p = p * weight
        a = a * weight
        b = b * weight
    x = b - a
    # projected length on the line
    t = np.dot(p - a, x) / np.dot(x, x)
    # orthogonal vector to the line
    n = (p - a) - np.outer(t, x)
    dist = np.linalg.norm(n, axis=1)
    return dist, t, n


def distance_to_refpop(
    sample: np.ndarray, anc1: np.ndarray, anc2: np.ndarray, weight: np.ndarray = None
):
    """
    Calculate the projection distance and position for every sample individuals to the
    line defined by two ancestry populations.

    Parameters
    ----------
    sample: np.ndarray
        (n_indiv, n_dim) matrix
    anc1: np.ndarray
        (n_indiv, n_dim) matrix
    anc2: np.ndarray
        (n_indiv, n_dim) matrix
    weight: np.ndarray, optional
        weight associated to each dimension, for example, could be sqrt(eigenvalues)
        if each dimension corresponds to a principal components.
        if not provided, equal weights for each dimension will be used

    Returns
    -------
    dist: np.ndarray
        normalized distance to the line defined by two ancestral populations
    t: np.ndarray
        projection position
    """
    n_dim = sample.shape[1]
    assert (anc1.shape[1] == n_dim) & (anc2.shape[1] == n_dim)
    if weight is not None:
        assert weight.ndim == 1
        assert len(weight) == n_dim
    p1, p2 = anc1.mean(axis=0), anc2.mean(axis=0)
    sample_dist, sample_t, sample_n = distance_to_line(sample, p1, p2, weight)
    anc1_dist, anc1_t, anc1_n = distance_to_line(anc1, p1, p2, weight)
    anc2_dist, anc2_t, anc2_n = distance_to_line(anc2, p1, p2, weight)
    anc1_maxdist, anc2_maxdist = np.max(anc1_dist), np.max(anc2_dist)
    sample_normalized_dist = sample_dist / (
        (1 - sample_t) * anc1_maxdist + sample_t * anc2_maxdist
    )
    return sample_normalized_dist, sample_t


import numpy as np
from scipy import stats
from scipy.special import logsumexp


def pval_to_posterior(pval):
    """
    Convert marginal p-value to posterior probability

    Args
    -----
    pval: (n_snp, )
    """
    zsc = stats.norm.ppf(pval / 2)
    logprob = -stats.norm.logpdf(zsc)
    return np.exp(logprob - logsumexp(logprob))


def posterior_to_credible_set(posterior, coverage=0.9):
    """
    From the given posterior probability to the credible set

    Args
    -----
    posterior: (n_snp, )
    coverage: desired coverage
    """
    n_cs = np.sum(np.cumsum(np.sort(posterior)[::-1]) < coverage) + 1
    return np.argsort(posterior)[::-1][0:n_cs]


def chi2_to_posterior(chi2, df):
    """
    Convert marginal p-value to the posterior probability of being causal

    Args
    -----
    pval: (n_snp, )
    df: degree of freedom
    """
    pval = stats.chi2.sf(chi2, df=df)
    return pval_to_posterior(pval)


from ._plot import (
    lanc,
    admixture,
    manhattan,
    qq,
    compare_pval,
    pca,
    joint_pca,
    rg_posterior,
    susie,
)


__all__ = ["lanc", "admixture", "manhattan", "qq", "susie"]



import numpy as np
from typing import List
import dask.array as da
import xarray as xr
from typing import Union, List, Dict
import admix
import pandas as pd
from tqdm import tqdm


def quant_pheno(
    dset: admix.Dataset,
    hsq: float,
    cor: float = None,
    n_causal: int = None,
    beta: np.ndarray = None,
    cov_cols: List[str] = None,
    cov_effects: List[float] = None,
    n_sim=10,
) -> dict:
    """Simulate continuous phenotype of admixed individuals [continuous]

    Parameters
    ----------
    dset: admix.Dataset
        Dataset containing the following variables:
            - geno: (n_indiv, n_snp, 2) phased genotype of each individual
            - lanc: (n_indiv, n_snp, 2) local ancestry of each SNP
    hsq: float
        Variance explained by the genotype effect
    cor: float
        Correlation between the genetic effects from two ancestral backgrounds
    n_causal: int, optional
        number of causal variables, by default None
    beta: np.ndarray, optional
        causal effect of each causal variable, by default None
    cov_cols: List[str], optional
        list of covariates to include as covariates, by default None
    cov_effects: List[float], optional
        list of the effect of each covariate, by default None
        for each simulation, the cov_effects will be the same
    n_sim : int, optional
        number of simulations, by default 10


    Returns
    -------
    beta: np.ndarray
        simulated effect sizes (2 * n_snp, n_sim)
    phe_g: np.ndarray
        simulated genetic component of phenotypes (n_indiv, n_sim)
    phe: np.ndarray
        simulated phenotype (n_indiv, n_sim)
    """
    n_anc = dset.n_anc

    apa = dset.allele_per_anc()
    n_snp, n_indiv = apa.shape[0:2]

    # simulate effect sizes
    if beta is None:
        if n_causal is None:
            # n_causal = n_snp if `n_causal` is not specified
            n_causal = n_snp

        if cor is None:
            cor = 1

        # if `beta` is not specified, simulate effect sizes
        beta = np.zeros((n_snp, n_anc, n_sim))

        # construct effect correlation matrix
        beta_cov = np.eye(n_anc)
        beta_cov[~np.eye(n_anc, dtype=bool)] = cor
        for i_sim in range(n_sim):
            cau = sorted(
                np.random.choice(np.arange(n_snp), size=n_causal, replace=False)
            )

            i_beta = np.random.multivariate_normal(
                mean=np.repeat(0, n_anc),
                cov=beta_cov,
                size=n_causal,
            )

            for i_anc in range(n_anc):
                beta[cau, i_anc, i_sim] = i_beta[:, i_anc]
    else:
        assert (cor is None) and (
            n_causal is None
        ), "If `beta` is specified, `var_g`, `var_e`, `gamma`, and `n_causal` must be specified"
        assert beta.shape == (n_snp, n_anc) or beta.shape == (
            n_snp,
            n_anc,
            n_sim,
        ), "`beta` must be of shape (n_snp, n_anc) or (n_snp, n_anc, n_sim)"
        if beta.shape == (n_snp, n_anc):
            # replicate `beta` for each simulation
            beta = np.repeat(beta[:, :, np.newaxis], n_sim, axis=2)

    pheno_g = np.zeros([n_indiv, n_sim])
    snp_chunks = apa.chunks[0]
    indices = np.insert(np.cumsum(snp_chunks), 0, 0)

    for i in tqdm(range(len(indices) - 1), desc="admix.simulate.quant_pheno"):
        start, stop = indices[i], indices[i + 1]
        apa_chunk = apa[start:stop, :, :].compute()
        for i_anc in range(n_anc):
            pheno_g += np.dot(apa_chunk[:, :, i_anc].T, beta[start:stop, i_anc, :])

    # scale variance of pheno_g to hsq
    std_scale = np.sqrt(hsq / np.var(pheno_g, axis=0))
    pheno_g *= std_scale
    beta *= std_scale

    pheno_e = np.zeros(pheno_g.shape)
    for i_sim in range(n_sim):
        pheno_e[:, i_sim] = np.random.normal(
            loc=0.0, scale=np.sqrt(1 - hsq), size=n_indiv
        )

    pheno = pheno_g + pheno_e

    # if `cov_cols` are specified, add the covariates to the phenotype
    if cov_cols is not None:
        # if `cov_effects` are not set, set to random normal values
        if cov_effects is None:
            cov_effects = np.random.normal(size=len(cov_cols))  # type: ignore
        # add the covariates to the phenotype
        cov_values = np.zeros((n_indiv, len(cov_cols)))
        for i_cov, cov_col in enumerate(cov_cols):
            cov_values[:, i_cov] = dset.indiv[cov_col].values
        pheno += np.dot(cov_values, cov_effects).reshape((n_indiv, 1))

    return {
        "beta": beta,
        "pheno_g": pheno_g,
        "pheno": pheno,
    }


# TODO: check https://github.com/TalShor/SciLMM/blob/master/scilmm/Estimation/HE.py


def quant_pheno_grm(
    dset: admix.Dataset,
    grm: Union[str, List[str], dict],
    var: Dict[str, float],
    cov_cols: List[str] = None,
    cov_effects: List[float] = None,
    n_sim=10,
):
    """Simulate continuous phenotype of admixed individuals [continuous] using GRM

    Parameters
    ----------
    grm: str, list of str or dict
        column name(s) of GRM, or a dict of {name: grm}
        Don't include the identity matrix, the indentify matrix representing environemntal
        factor will be added to the list automatically
    var: dict of {str: float}
        dictionary of variance explained by the GRM effect, use 'e' to set the variance of
        environmental effectse.g. {'K1': 0.5, 'K2': 0.5, 'e': 1.0}
    gamma: float, optional
        Correlation between the genetic effects from two ancestral backgrounds, by default None
    cov_cols: List[str], optional
        list of covariates to include as covariates, by default None
    cov_effects: List[float], optional
        list of the effect of each covariate, by default None
        for each simulation, the cov_effects will be the same
    n_sim : int, optional
        number of simulations, by default 10

    Returns
    -------
    A dictionary containing the following variables:
        - pheno: (n_indiv, n_sim) simulated phenotype
        - cov_effects: (n_cov,) simulated covariate effects
    """
    # get the dictionary of all the GRMs (including environmental)
    # name -> (grm, var)
    if isinstance(grm, dict):
        # obtain from the list
        grm_var = {k: (grm[k], var[k]) for k in grm}
    elif isinstance(grm, list):
        grm_var = {k: (dset[k].data, var[k]) for k in grm}
    elif isinstance(grm, str):
        grm_var = {grm: (dset[grm].data, var[grm])}
    else:
        raise ValueError("`grm` must be a dictionary, list or string")

    # add environemntal effect
    n_indiv = dset.dims["indiv"]
    grm_var["e"] = (np.eye(n_indiv), var["e"])

    covariance = da.zeros((n_indiv, n_indiv))
    for name in grm_var:
        covariance += grm_var[name][0] * grm_var[name][1]

    # fixed effects
    # if `cov_cols` are specified, add the covariates to the phenotype
    if cov_cols is not None:
        # if `cov_effects` are not set, set to random normal values
        if cov_effects is None:
            cov_effects = np.random.normal(size=len(cov_cols))  # type: ignore
        cov_values = np.zeros((n_indiv, len(cov_cols)))
        for i_cov, cov_col in enumerate(cov_cols):
            cov_values[:, i_cov] = dset[cov_col + "@indiv"].values
        fixed_effects = np.dot(cov_values, cov_effects)
    else:
        fixed_effects = np.zeros(n_indiv)
    # simulate phenotype
    ys = np.random.multivariate_normal(
        mean=fixed_effects,
        cov=covariance,
        size=n_sim,
    ).T

    return {"pheno": ys, "cov_effects": cov_effects}


def sample_case_control(pheno: np.ndarray, control_ratio: float) -> np.ndarray:
    """Sample case control from the population with a desired ratio

    Args:
        pheno (np.ndarray): (n_indiv, ) binary vector representing the case control status
            for each individual
        control_ratio (float): the ratio of control / case

    Returns:
        np.ndarray: (n_indiv, ) vector indicating whether i-th individual is sampled
    """
    case_index = np.where(pheno == 1)[0]
    control_index = np.random.choice(
        np.where(pheno == 0)[0],
        size=int(len(case_index) * control_ratio),
        replace=False,
    )
    study_index = np.sort(np.concatenate([case_index, control_index]))
    return study_index


def binary_pheno(
    dset: admix.Dataset,
    hsq: float = 0.5,
    cor: float = None,
    n_causal: int = None,
    case_prevalence: float = 0.1,
    beta: np.ndarray = None,
    cov_cols: List[str] = None,
    cov_effects: List[float] = None,
    n_sim=10,
    method="probit",
):
    """
    Simulate under liability threshold. First simulate quantative traits in the liability
    scale. With the given `case_prevalence`, convert liability to binary traits.


    Parameters
    ----------
    hsq: if method is "probit", this corresponds to the proportion of variance explained
    by the genotype in the liability scale.

    if method is "logit", this corresponds to the variance explained of (X\beta),
    and hsq can be larger than 1, in the case.
    """

    from scipy import stats

    assert method in ["probit", "logit"]

    # warn if "logit" is used
    if method == "logit":
        admix.logger.warn(
            "`logit` method for simulating binary phenotype is not well tested."
            "Consider using `probit` method instead or this function with caution."
        )

    # build a skeleton of phenotype using the code from quant_pheno

    if method == "probit":
        quant_sim = quant_pheno(
            dset=dset,
            hsq=hsq,
            cor=cor,
            n_causal=n_causal,
            beta=beta,
            cov_cols=cov_cols,
            cov_effects=cov_effects,
            n_sim=n_sim,
        )

        beta = quant_sim["beta"]
        pheno_g = quant_sim["pheno_g"]
        qpheno = quant_sim["pheno"]
        qpheno -= qpheno.mean(axis=0)
        case_threshold = -stats.norm.ppf(case_prevalence)

        pheno = np.zeros_like(qpheno)
        pheno[qpheno < case_threshold] = 0
        pheno[qpheno >= case_threshold] = 1

        return {
            "beta": beta,
            "pheno_g": pheno_g,
            "pheno": pheno.astype(int),
        }

    elif method == "logit":
        from scipy.optimize import fsolve
        from scipy.special import logit, expit

        quant_sim = quant_pheno(
            dset=dset,
            hsq=0.5,
            cor=cor,
            n_causal=n_causal,
            beta=beta,
            cov_cols=cov_cols,
            cov_effects=cov_effects,
            n_sim=n_sim,
        )
        beta = quant_sim["beta"]
        pheno_g = quant_sim["pheno_g"]

        # scale variance of pheno_g to hsq
        std_scale = np.sqrt(hsq / np.var(pheno_g, axis=0))
        pheno_g *= std_scale
        beta *= std_scale
        pheno = np.zeros_like(pheno_g)
        for sim_i in range(n_sim):
            # find an intercept, such that the expectation is case_prevalence.
            func = lambda b: np.mean(expit(b + pheno_g[:, sim_i])) - case_prevalence
            intercept = fsolve(func, logit(case_prevalence))
            pheno[:, sim_i] = np.random.binomial(
                1, expit(intercept + pheno_g[:, sim_i])
            )
        return {
            "beta": beta,
            "pheno_g": pheno_g,
            "pheno": pheno.astype(int),
        }
    else:
        raise NotImplementedError


def quant_pheno_1pop(
    geno: da.Array,
    hsq: float,
    n_causal: int = None,
    beta: np.ndarray = None,
    snp_prior_var: np.ndarray = None,
    cov: np.ndarray = None,
    cov_effects: List[float] = None,
    n_sim=10,
) -> dict:
    """Simulate quantative phenotype for a single population [continuous]

    Parameters
    ----------
    geno: da.Array
        (n_snp, n_indiv) array of genotype
    hsq: float
        Proportion of variance explained by the genotype effects
    n_causal: int, optional
        number of causal variables, by default None
    beta: np.ndarray, optional
        Effect sizes
    snp_prior_var: np.ndarray, optional
        Prior variance of each SNP
    cov_cols: List[str], optional
        list of covariates to include as covariates, by default None
    cov_effects: List[float], optional
        list of the effect of each covariate, by default None
        for each simulation, the cov_effects will be the same
    n_sim : int, optional
        number of simulations, by default 10

    Returns
    -------
    beta: np.ndarray
        simulated effect sizes (n_snp, n_sim)
    phe_g: np.ndarray
        simulated genetic component of phenotypes (n_indiv, n_sim)
    phe: np.ndarray
        simulated phenotype (n_indiv, n_sim)
    """
    n_snp, n_indiv = geno.shape

    # simulate effect sizes
    if beta is None:
        assert n_causal is not None, "n_causal must be specified if beta is None"

        if snp_prior_var is None:
            snp_prior_var = np.ones(n_snp)

        assert n_causal <= n_snp, "n_causal must be <= n_snp"
        # if `beta` is not specified, simulate effect sizes
        beta = np.zeros((n_snp, n_sim))
        for i_sim in range(n_sim):
            cau = sorted(
                np.random.choice(np.arange(n_snp), size=n_causal, replace=False)
            )
            i_beta = np.random.normal(
                loc=0.0,
                scale=1.0,
                size=n_causal,
            )
            i_beta = i_beta * np.sqrt(snp_prior_var[cau])
            beta[cau, i_sim] = i_beta
    else:
        assert (
            n_causal is None
        ), "If `beta` is specified, `var_g`, `var_e`, and `n_causal` must be specified"
        assert beta.shape == (n_snp,) or beta.shape == (
            n_snp,
            n_sim,
        ), "`beta` must be of shape (n_snp,) or (n_snp, n_sim)"
        if beta.shape == (n_snp,):
            # replicate `beta` for each simulation
            beta = np.repeat(beta[:, np.newaxis], n_sim, axis=1)
            print(beta.shape)

    pheno_g = admix.data.geno_mult_mat(geno, beta, mat_dim="snp")

    # standardize the phe_g so that the variance of g is `var_g`
    std_scale = np.sqrt(hsq / np.var(pheno_g, axis=0))
    pheno_g *= std_scale
    beta *= std_scale

    assert 0 <= hsq <= 1, "hsq must be between 0 and 1"
    pheno_e = np.zeros(pheno_g.shape)
    for i_sim in range(n_sim):
        pheno_e[:, i_sim] = np.random.normal(
            loc=0.0, scale=np.sqrt(1 - hsq), size=n_indiv
        )

    pheno = pheno_g + pheno_e

    # if `cov_cols` are specified, add the covariates to the phenotype
    if cov is not None:
        # if `cov_effects` are not set, set to random normal values
        if cov_effects is None:
            cov_effects = np.random.normal(size=cov.shape[0])  # type: ignore
        # add the covariates to the phenotype
        pheno += np.dot(cov, cov_effects).reshape((n_indiv, 1))

    return {
        "beta": beta,
        "pheno_g": pheno_g,
        "pheno": pheno,
        "cov_effects": cov_effects,
    }



import numpy as np
from typing import List
import dask.array as da
import admix
import pandas as pd
from tqdm import tqdm
from ._lanc import hap_lanc as simulate_hap_lanc


def admix_geno(
    geno_list: List[da.Array],
    df_snp: pd.DataFrame,
    anc_props: List[float],
    mosaic_size: float,
    n_indiv: int,
    return_sparse_lanc=False,
) -> admix.Dataset:
    """Simulate admixed genotype

    The generative model is:

    - for each ancestry, the allele frequencies are drawn
    - for each individual, breakpoints are drawn from a Poisson process. and the ancestry
        will be filled based a multinomial distribution with `n_anc` components
    - for each SNP, given the ancestry and the allele frequencies, the haplotype is drawn.
        Haplotype are simulated under some frequencies

    Parameters
    ----------
    geno_list : List[da.Array]
        List of ancestral data sets, each with (n_snp, n_indiv)
    df_snp : pd.DataFrame
        Dataframe of SNPs shared across ancestral data sets
    anc_props : list of float
        Proportion of ancestral populations
    mosaic_size : float
        Expected mosaic size in # of SNPs. use admix.lanc.calculate_mosaic_size() to
        calculate the mosaic size
    n_indiv : int
        Number of individuals to simulate

    Returns
    -------
    admix.Dataset
        Simulated admixed dataset
    """

    n_anc = len(geno_list)
    n_snp = geno_list[0].shape[0]
    assert all(
        n_snp == geno.shape[0] for geno in geno_list
    ), "all geno must have the same number of SNPs"
    assert n_snp == df_snp.shape[0], "df_snp must have the same number of SNPs"

    assert np.sum(anc_props) == 1, "anc_props must sum to 1"
    assert len(anc_props) == n_anc, "anc_props must have the same length as n_anc"
    anc_props = np.array(anc_props)

    dset_hap_list = [
        da.hstack([geno[:, :, 0], geno[:, :, 1]]).compute() for geno in geno_list
    ]

    hap_lanc_breaks, hap_lanc_values = admix.simulate.hap_lanc(
        n_snp=n_snp, n_hap=n_indiv * 2, mosaic_size=mosaic_size, anc_props=anc_props
    )
    geno = np.zeros((n_snp, n_indiv * 2), dtype=np.int8)

    for hap_i in tqdm(range(n_indiv * 2)):
        start = 0
        for stop, val in zip(hap_lanc_breaks[hap_i], hap_lanc_values[hap_i]):
            geno[start:stop, hap_i] = dset_hap_list[val][
                start:stop, np.random.randint(dset_hap_list[val].shape[1])
            ]
            start = stop
    lanc_breaks, lanc_values = admix.data.haplo2diplo(
        breaks=hap_lanc_breaks, values=hap_lanc_values
    )
    lanc = admix.data.Lanc(breaks=lanc_breaks, values=lanc_values)

    # a = np.random.randn(4, 5, 2)
    # b = np.zeros((4, 10))
    # b[:, 0::2] = a[:, :, 0]
    # b[:, 1::2] = a[:, :, 1]
    # assert np.all(b == a.reshape(4, 10))
    # assert np.all(b.reshape(4, 5, 2) == a)

    # equivalent: geno = np.dstack([geno[:, 0::2], geno[:, 1::2]])
    geno = geno.reshape(n_snp, n_indiv, 2)

    dset = admix.Dataset(
        geno=da.from_array(geno, chunks=-1),
        lanc=lanc.dask(snp_chunk=n_snp),
        n_anc=n_anc,
        snp=df_snp,
        indiv=pd.DataFrame(
            {"indiv": ["indiv_" + str(i) for i in np.arange(n_indiv)]}
        ).set_index("indiv"),
    )
    if return_sparse_lanc:
        return dset, lanc
    else:
        return dset


def admix_geno_simple(
    n_indiv: int,
    n_snp: int,
    n_anc: int,
    mosaic_size: float,
    anc_props: np.ndarray,
    allele_freqs: List[np.ndarray] = None,
    af_maf_low: float = 0.05,
    af_maf_high: float = 0.5,
) -> admix.Dataset:
    """Simulate admixed genotype

    The generative model is:

    - for each ancestry, the allele frequencies are drawn
    - for each individual, breakpoints are drawn from a Poisson process. and the ancestry will be filled based a multinomial distribution with `n_anc` components
    - for each SNP, given the ancestry and the allele frequencies, the haplotype is drawn. Haplotype are simulated under some frequencies

    Parameters
    ----------
    n_indiv : int
        Number of individuals
    n_snp : int
        Number of SNPs
    n_anc : int
        Number of ancestries
    mosaic_size : float
        Expected mosaic size in # of SNPs
    anc_props : list of float
        Proportion of ancestral populations, if not specified, the proportion
        is uniform over the ancestral populations.
    allele_freqs: List[np.ndarray]
        Allele frequencies, if not specified, the frequencies are drawn from
        using `af_maf_low` and `af_maf_high`
    af_maf_low : float
        Lowest allowed minor allele frequency
    af_maf_low : float
        Highest allowed minor allele frequency

    Returns
    -------
    admix.Dataset
        Simulated admixed dataset


    References
    ----------
    1. https://github.com/slowkoni/rfmix/blob/master/simulate.cpp
    2. https://github.com/slowkoni/admixture-simulation
    3. https://github.com/williamslab/admix-simu
    """
    if anc_props is None:
        anc_props = np.ones(n_anc) / n_anc
    else:
        anc_props = np.array(anc_props)
        assert np.sum(anc_props) == 1, "anc_props must sum to 1"
    anc_props = np.array(anc_props)
    assert anc_props.size == n_anc, "anc_props must have the same length as n_anc"

    breaks, values = simulate_hap_lanc(
        n_hap=n_indiv * 2, n_snp=n_snp, mosaic_size=mosaic_size, anc_props=anc_props
    )
    breaks, values = admix.data.haplo2diplo(breaks=breaks, values=values)
    rls_lanc = (
        admix.data.Lanc(breaks=breaks, values=values).dask().swapaxes(0, 1).compute()
    )
    # n_indiv x n_snp x 2 (2 for each haplotype)
    # rls_lanc = np.dstack([rls_lanc[0:n_indiv, :], rls_lanc[n_indiv:, :]])
    rls_geno = np.zeros_like(rls_lanc)
    if allele_freqs is None:
        # allele frequencies for the two populations
        allele_freqs = [
            np.random.uniform(low=af_maf_low, high=af_maf_high, size=n_snp)
            for _ in range(n_anc)
        ]
    else:
        assert (
            len(allele_freqs) == n_anc
        ), "allele_freqs must have the same length as n_anc"
        assert np.all(
            [len(af) == n_snp for af in allele_freqs]
        ), "each element in allele_freqs must have the same length as n_snp"

    for i_anc in range(n_anc):
        rls_geno[rls_lanc == i_anc] = np.random.binomial(
            n=1, p=allele_freqs[i_anc][np.where(rls_lanc == i_anc)[1]]
        )

    return admix.Dataset(
        geno=da.from_array(np.swapaxes(rls_geno, 0, 1), chunks=-1),
        lanc=da.from_array(np.swapaxes(rls_lanc, 0, 1), chunks=-1),
        n_anc=n_anc,
    )


def admix_geno2(
    n_indiv: int,
    n_snp: int,
    n_anc: int,
    mosaic_size: float,
    anc_props: np.ndarray = None,
    ancestral_af_range: List[float] = [0.1, 0.9],
    F_st: float = 0.3,
) -> admix.Dataset:
    """Simulate admixed genotype
    The generative model is:
        1. for each ancestry, the allele frequencies are drawn
        2. for each individual, breakpoints are drawn from a Poisson process.
            and the ancestry will be filled based a multinomial distribution with `n_anc`
            components
        3. for each SNP, given the ancestry and the allele frequencies, the haplotype
            is drawn
    Haplotype are simulated under some frequencies

    Parameters
    ----------
    n_indiv : int
        Number of individuals
    n_snp : int
        Number of SNPs
    n_anc : int
        Number of ancestries
    mosaic_size : float
        Expected mosaic size in # of SNPs
    anc_props : list of float
        Proportion of ancestral populations, if not specified, the proportion
        is uniform over the ancestral populations.
    ancestral_af_range: [float, float]
        allele frequencies range.
    F_st: float
        Distance for the two populations that form the admixed population.

    Returns
    -------
        admix.Dataset
        Simulated admixed dataset

    References
    ----------
    We follow Zaitlen et al. 2014, Nature Genetics
    See description in subsection "Simulations with simulated genotypes" in Methods
    """
    assert n_anc == 2, "n_anc must be 2, only two-way admixture is currently supported"
    # global ancestry
    if anc_props is None:
        anc_props = np.ones(n_anc) / n_anc
    else:
        anc_props = np.array(anc_props)
        assert np.sum(anc_props) == 1, "anc_props must sum to 1"
    anc_props = np.array(anc_props)
    assert anc_props.size == n_anc, "anc_props must have the same length as n_anc"

    # local ancestry
    rls_lanc = (
        simulate_hap_lanc(
            n_hap=n_indiv * 2, n_snp=n_snp, mosaic_size=mosaic_size, anc_props=anc_props
        )
        .dask()
        .T
    )
    # n_indiv x n_snp x 2 (2 for each haplotype)
    rls_lanc = np.dstack([rls_lanc[0:n_indiv, :], rls_lanc[n_indiv:, :]])
    rls_geno = np.zeros_like(rls_lanc)

    ancestral_afs = np.random.uniform(
        low=ancestral_af_range[0], high=ancestral_af_range[1], size=n_snp
    )

    # allele frequencies for the two populations
    allele_freqs = [
        np.random.beta(
            a=ancestral_afs * (1 - F_st) / F_st,
            b=(1 - ancestral_afs) * (1 - F_st) / F_st,
        )
        for _ in range(n_anc)
    ]

    for i_anc in range(n_anc):
        rls_geno[rls_lanc == i_anc] = np.random.binomial(
            n=1, p=allele_freqs[i_anc][np.where(rls_lanc == i_anc)[1]]
        )
    return admix.Dataset(
        geno=da.from_array(np.swapaxes(rls_geno, 0, 1)),
        lanc=da.from_array(np.swapaxes(rls_lanc, 0, 1)),
        n_anc=n_anc,
    )



import pandas as pd
import numpy as np
import admix
from typing import (
    List,
    Union,
    Tuple,
    Sequence,
)


def all_array_equal(array_list: List[np.ndarray]) -> bool:
    """check whether all arrays in the list are equal

    Parameters
    ----------
    List[np.ndarray]

    Returns
    -------
    bool
    """
    return all([np.allclose(array_list[0], a) for a in array_list[1:]])


def align_datasets(dsets: List[admix.Dataset], dim: str) -> List[admix.Dataset]:
    """takes 2 or more datasets, return the aligned dataset

    SNP alignment is based on index, CHROM, POS, REF, ALT
    Individual alignment is based on index

    TODO: refer to PLINK2 alignment page for preprocessing your data to align SNP attributes
    before hand.

    Parameters
    ----------
    dsets : List[Dataset]
        List of datasets
    dim : str
        which dimension to check

    Returns
    -------
    List[admix.Dataset]: list of aligned datasets
    """
    assert dim in ["snp", "indiv"]
    if dim == "snp":
        required_snp_columns = admix.dataset.REQUIRED_SNP_COLUMNS
        # TODO: find aligned snp index among datasets


def check_align(dsets: List[admix.Dataset], dim: str) -> bool:
    """takes 2 or more datasets, and check whether the common attributes among datasets
    are properly aligned

    Parameters
    ----------
    dsets : List[admix.Dataset]
        List of datasets
    dim : str
        which dimension to check

    Returns
    -------
    bool: whether the two datasets align in the given dimension
    """
    aligned = False
    assert dim in ["snp", "indiv"], "dim must be either 'snp' or 'indiv'"
    if dim == "snp":
        required_snp_columns = admix.dataset.REQUIRED_SNP_COLUMNS

        # find common columns among datasets
        common_cols = set.intersection(*[set(dset.snp.columns) for dset in dsets])

        # common_cols must be a subset of required_snp_columns
        assert set(required_snp_columns).issubset(
            common_cols
        ), "common_cols must be a subset of required_snp_columns"

        inconsistent_cols = [
            col
            for col in common_cols
            if not all_array_equal([dset.snp[col].values for dset in dsets])
        ]
        if not all_array_equal([dset.snp.index.values for dset in dsets]):
            inconsistent_cols.append("index")
        if len(inconsistent_cols) > 0:
            print(f"inconsistent columns: {','.join(inconsistent_cols)} in snp")
            aligned = False
        else:
            aligned = True

    elif dim == "indiv":
        # find common columns among datasets
        common_cols = set.intersection(*[set(dset.indiv.columns) for dset in dsets])
        # check whether all datasets have the same columns
        inconsistent_cols = [
            col
            for col in common_cols
            if not all_array_equal([dset.indiv[col].values for dset in dsets])
        ]
        if len(inconsistent_cols) > 0:
            print(f"inconsistent columns: {','.join(inconsistent_cols)} in indiv")
            aligned = False
        else:
            aligned = True
    return aligned



import admix
import pandas as pd
import numpy as np
from tqdm import tqdm
from typing import List, Dict
import glob
from natsort import natsorted
import os
import json
from scipy import stats
from ._utils import log_params
from scipy.interpolate import CubicSpline


def _write_admix_grm(
    dict_grm: Dict[str, np.ndarray],
    df_weight: pd.Series,
    df_id: pd.DataFrame,
    out_prefix: str,
):
    """write admix grm

    Parameters
    ----------
    dict_grm : Dict[str, np.ndarray]
        name -> grm matrix
    df_weight : pd.Series
        snp weights
    n_snp : int
        number of snps
    df_id: pd.DataFrame
        list of individuals
    out_prefix: str
        prefix of the output file

    Returns
    -------
    None
    """
    for name in dict_grm:
        grm = dict_grm[name]
        admix.tools.gcta.write_grm(
            out_prefix + "." + name,
            K=grm,
            df_id=df_id,
            n_snps=np.repeat(len(df_weight), len(df_id)),
        )

    # write weight
    df_weight.to_csv(out_prefix + ".weight.tsv", sep="\t")


# Implementing genetic correlation related functions
def admix_grm(
    pfile: str,
    out_prefix: str,
    maf_cutoff: float = 0.005,
    her_model="mafukb",
    freq_cols=["LANC_FREQ1", "LANC_FREQ2"],
    snp_chunk_size: int = 256,
    snp_list: str = None,
    write_raw: bool = False,
) -> None:
    """
    Calculate the admix GRM for a given pfile

    Parameters
    ----------
    pfile : str
        Path to the pfile
    out_prefix : str
        Prefix of the output files
    maf_cutoff : float, optional
        MAF cutoff for the admixed individuals, by default 0.005
    her_model : str, optional
        Heritability model, by default "mafukb"
        one of "uniform", "gcta", "ldak", "mafukb"
    freq_cols : List[str], optional
        Columns of the pfile to use as frequency, by default ["LANC_FREQ1", "LANC_FREQ2"]
        to perform the ancestry-specific MAF cutoffs
    snp_chunk_size : int, optional
        Number of SNPs to read at a time, by default 256
        This can be tuned to reduce memory usage
    snp_list : str, optional
        Path to a file containing a list of SNPs to use. Each line should be a SNP ID.
        Only SNPs in the list will be used for the analysis. By default None
    write_raw: bool, optional
        Whether to write the raw GRM, G1, G2, G12, by default False
    Returns
    -------
    GRM files: {out_prefix}.[K1, K2].[grm.bin | grm.id | grm.n] will be generated
    Weight file: {out_prefix}.weight.tsv will be generated
    """

    log_params("admix-grm", locals())
    assert len(freq_cols) == 2, "freq_cols must be a list of length 2"
    dset = admix.io.read_dataset(pfile=pfile, snp_chunk=snp_chunk_size)

    # filter for SNPs
    if snp_list is not None:
        with open(snp_list, "r") as f:
            filter_snp_list = [line.strip() for line in f]
        n_filter_snp = len(filter_snp_list)
        filter_snp_list = dset.snp.index[dset.snp.index.isin(filter_snp_list)]
        if len(filter_snp_list) < n_filter_snp:
            admix.logger.warning(
                f"{n_filter_snp - len(filter_snp_list)} SNPs in {snp_list} are not in the dataset"
            )
        dset = dset[filter_snp_list]

    snp_subset = np.where(
        dset.snp[freq_cols[0]].between(maf_cutoff, 1 - maf_cutoff)
        & dset.snp[freq_cols[1]].between(maf_cutoff, 1 - maf_cutoff)
    )[0]

    dset = dset[snp_subset]

    dset.snp["PRIOR_VAR"] = admix.data.calc_snp_prior_var(dset.snp, her_model=her_model)

    K1, K2 = admix.data.admix_grm_equal_var(
        geno=dset.geno,
        lanc=dset.lanc,
        snp_prior_var=dset.snp.PRIOR_VAR.values,
        n_anc=dset.n_anc,
    )

    df_weight = dset.snp.PRIOR_VAR
    df_id = pd.DataFrame({"0": dset.indiv.index.values, "1": dset.indiv.index.values})
    _write_admix_grm(
        dict_grm={"K1": K1, "K2": K2},
        df_weight=df_weight,
        df_id=df_id,
        out_prefix=out_prefix,
    )

    if write_raw:
        G1, G2, G12 = admix.data.admix_grm(
            geno=dset.geno,
            lanc=dset.lanc,
            snp_prior_var=dset.snp.PRIOR_VAR.values,
        )

        for suffix, mat in zip(["G1", "G2", "G12"], [G1, G2, G12]):
            admix.tools.gcta.write_grm(
                f"{out_prefix}.{suffix}",
                K=mat,
                df_id=df_id,
                n_snps=np.repeat(len(df_weight), len(df_id)),
            )


def admix_grm_merge(prefix: str, out_prefix: str, n_part: int = 22) -> None:
    """
    Merge multiple GRM matrices

    Parameters
    ----------
    prefix : str
        Prefix of the GRM files, any files with the pattern of <prefix>.*
        will be merged
    out_prefix : str
        Prefix of the output file
    n_part : int, optional
        Number of partitions, by default 22

    Returns
    -------
    GRM files: {out_prefix}.[K1, K2].[grm.bin | grm.id | grm.n] will be generated
    Weight file: {out_prefix}.weight.tsv will be generated
    """
    log_params("admix-grm-merge", locals())

    # search for files with the pattern of <prefix>.<suffix>.K1.grm.bin and <prefix>.<suffix>.K2.grm.bin
    K1_grm_prefix_list = [
        f[: -len(".K1.grm.bin")] for f in sorted(glob.glob(f"{prefix}*.K1.grm.bin"))
    ]
    K2_grm_prefix_list = [
        f[: -len(".K2.grm.bin")] for f in sorted(glob.glob(f"{prefix}*.K2.grm.bin"))
    ]
    assert len(K1_grm_prefix_list) == len(K2_grm_prefix_list)
    assert K1_grm_prefix_list == K2_grm_prefix_list, (
        "GRM files .K1 and .K2 are not matched, "
        f"K1_grm_list={K1_grm_prefix_list}, K2_grm_list={K2_grm_prefix_list}"
    )
    grm_prefix_list = natsorted(K1_grm_prefix_list)
    if len(grm_prefix_list) != n_part:
        raise ValueError(
            f"Number of GRM files ({len(grm_prefix_list)}) is not equal to n_part ({n_part})"
        )
    admix.logger.info(f"{len(grm_prefix_list)} GRM files to be merged: {grm_prefix_list}")
    prior_var_list = []
    for grm_prefix in grm_prefix_list:
        prior_var_list.append(pd.read_csv(grm_prefix + ".weight.tsv", sep="\t", index_col=0))

    def _merge(suffix):
        total_grm = 0
        total_df_id = None

        weight_list = []
        for i, grm_prefix in enumerate(grm_prefix_list):
            prior_var = prior_var_list[i]
            weight = prior_var["PRIOR_VAR"].sum()
            weight_list.append(weight)
            grm, df_id, n_snps = admix.tools.gcta.read_grm(f"{grm_prefix}.{suffix}")
            if total_df_id is None:
                total_df_id = df_id
            else:
                assert np.all(
                    total_df_id == df_id
                ), f"df_id of {grm_prefix} is not matched does not match with already loaded ones"
            total_grm += grm * weight

            assert np.all(n_snps == prior_var.shape[0])
        total_grm /= np.sum(weight_list)
        return total_grm, df_id

    K1, df_id1 = _merge("K1")
    K2, df_id2 = _merge("K2")
    assert df_id1.equals(df_id2)
    df_id = df_id1
    df_weight = pd.concat(prior_var_list)

    _write_admix_grm(
        dict_grm={"K1": K1, "K2": K2},
        df_weight=df_weight,
        df_id=df_id,
        out_prefix=out_prefix,
    )


def genet_cor(
    pheno: str,
    grm_prefix: str,
    out_dir: str,
    rg_grid=np.linspace(0, 1.0, 21),
    quantile_normalize: bool = True,
    n_thread: int = 2,
    clean: bool = True,
):
    """Estimate genetic correlation

    Parameters
    ----------
    pheno : str
        phenotype file, the 1st column contains ID, 2nd column contains phenotype, and
        the rest of columns are covariates.
    grm_prefix : str
        folder containing K1, K2 GRM files
    out_dir : str
        folder to store the output files
    rg_grid : list, optional
        List of rg values to grid search, by default np.linspace(0, 1.0, 21)
    quantile_normalize: bool
        whether to perform quantile normalization for both phenotype and each column of covariates
    n_thread : int, optional
        number of threads, by default 2
    """
    log_params("genet-cor", locals())

    ## compile phenotype and covariates
    df_pheno = pd.read_csv(pheno, sep="\t", index_col=0)
    df_pheno.index = df_pheno.index.astype(str)

    # subset for individuals with non-nan value in df_trait
    trait_col = df_pheno.columns[0]
    covar_cols = df_pheno.columns[1:]

    # filter out individuals with missing phenotype
    df_pheno = df_pheno[df_pheno[trait_col].notna()]

    df_trait = df_pheno[[trait_col]].copy()
    df_covar = df_pheno[covar_cols].copy()
    df_covar = admix.data.convert_dummy(df_covar)
    if quantile_normalize:
        # perform quantile normalization
        for col in df_trait.columns:
            df_trait[col] = admix.data.quantile_normalize(df_trait[col])

        for col in df_covar.columns:
            df_covar[col] = admix.data.quantile_normalize(df_covar[col])

    # fill na with column mean
    df_covar.fillna(df_covar.mean(), inplace=True)

    df_id = pd.DataFrame(
        {"FID": df_trait.index.values, "IID": df_trait.index.values},
        index=df_trait.index.values,
    )
    df_trait = pd.merge(df_id, df_trait, left_index=True, right_index=True)
    df_covar = pd.merge(df_id, df_covar, left_index=True, right_index=True)

    ## load grm
    K1, df_id1, n_snps1 = admix.tools.gcta.read_grm(grm_prefix + ".K1")
    K2, df_id2, n_snps2 = admix.tools.gcta.read_grm(grm_prefix + ".K2")
    assert df_id1.equals(df_id2)
    assert np.allclose(n_snps1, n_snps2)
    df_id = df_id1
    n_snps = n_snps1

    os.makedirs(out_dir, exist_ok=True)
    for rg in tqdm(rg_grid):
        K = K1 + K2 * rg

        grm = os.path.join(out_dir, f"rg{int(rg * 100)}")
        admix.tools.gcta.write_grm(
            grm,
            K=K,
            df_id=df_id,
            n_snps=n_snps,
        )
        admix.tools.gcta.reml(
            grm_path=grm,
            df_pheno=df_trait,
            df_covar=df_covar,
            out_prefix=os.path.join(out_dir, f"rg{int(rg * 100)}"),
            n_thread=n_thread,
            est_fix=True,
        )
        if clean:
            # remove <grm>.grm.* files
            for f in glob.glob(grm + ".grm.*"):
                os.remove(f)


def summarize_genet_cor(
    est_dir: str,
    out_prefix: str,
    weight_file: str = None,
    freq_file: str = None,
    scale_factor: float = None,
    freq_col: str = "FREQ",
    index_col: str = "snp",
    rg_str: str = "rg",
):
    """Summarize the results of genetic correlation analysis.

    Parameters
    ----------
    est_dir : str
        Estimation directory, containing rho<rho>.hsq, rho<rho>.log
    out_prefix : str
        output prefix
    weight_file: str
        weight_file specifying the prior variance file (<grm_prefix>.weight.tsv),
    freq_file: str
        frequency file (dataset *.snp_info files)
    scale_factor: float
        rather calculating the scale factor from `weight_file` and `freq_file` from
        scratch, specify the scale factor. This scale factor be pre-computed from
        admix.tools.gcta.calculate_hsq_scale
    freq_col: str
        column name for frequency in freq_file
    index_col: str
        column name for index in freq_file
    rg_str : str
        string name for rg, by default "rg", or "rho" (for legacy)

    Returns
    -------
    Log-likelihood curve for different rho: <out_prefix>.loglkl.txt
    Summarization file: <out_prefix>.summary.json. This file contains
        - poterior mode
        - highest posterior density interval (50% / 95%)
        - heritability (if grm_prefix is provided)

    Notes
    -----
    If `weight_file` and `freq_file` are provided, heritability at rg = 1
    (using rho100.hsq) will be estimated.
    """

    log_params("summarize-genet-cor", locals())

    rg_list = np.array(
        sorted(
            [
                int(os.path.basename(p).split(".")[0][len(rg_str) :])
                for p in glob.glob(os.path.join(est_dir, f"{rg_str}*.hsq"))
            ]
        )
    )
    admix.logger.info(f"rg={rg_list}")
    # read log-likelihood curve
    n_indiv = None
    loglkl_list = []
    for rg in rg_list:
        dict_reml = admix.tools.gcta.read_reml(os.path.join(est_dir, f"{rg_str}{rg}"))
        if n_indiv is None:
            n_indiv = dict_reml["n"]
        else:
            assert (
                n_indiv == dict_reml["n"]
            ), f"n_indiv={dict_reml['n']} from r={rg} different from previous one {n_indiv}"
        loglkl_list.append(dict_reml["loglik"])

    # interpolate
    rg_list = rg_list / 100

    # write raw estimation file
    pd.DataFrame({"rg": rg_list, "loglkl": loglkl_list}).to_csv(
        out_prefix + ".loglkl.txt", sep="\t", index=False
    )
    admix.logger.info(f"Log-likehood curves written to {out_prefix}.loglkl.txt")

    # summarize results
    assert rg_list[-1] == 1, "r=1 (r.hsq) should be included"
    dense_rg_list = np.linspace(min(rg_list), max(rg_list), 1001)
    dense_loglkl_list = CubicSpline(rg_list, loglkl_list)(dense_rg_list)

    dict_summary = {
        "n": n_indiv,
        "rg_mode": dense_rg_list[dense_loglkl_list.argmax()],
        "rg_hpdi(50%)": admix.data.hdi(dense_rg_list, dense_loglkl_list, ci=0.5),
        "rg_hpdi(95%)": admix.data.hdi(dense_rg_list, dense_loglkl_list, ci=0.95),
        "rg=1_pval": stats.chi2.sf((dense_loglkl_list.max() - dense_loglkl_list[-1]) * 2, df=1),
    }

    assert (weight_file is None) == (
        freq_file is None
    ), "weight_file and freq_file should be provided together"

    if (weight_file is not None) and (freq_file is not None):
        scale_factor = admix.tools.gcta.calculate_hsq_scale(
            weight_file=weight_file,
            freq_file=freq_file,
            freq_col=freq_col,
            index_col=index_col,
        )
        admix.logger.info(f"Computed hsq scale factor = {scale_factor:.3g}")

    if scale_factor is not None:
        dict_reml = admix.tools.gcta.read_reml(os.path.join(est_dir, f"{rg_str}100"))
        est_hsq, est_hsq_var = admix.tools.gcta.estimate_hsq(dict_reml, scale_factor=scale_factor)
        est_hsq_stderr = np.sqrt(est_hsq_var)
        dict_summary["hsq_est"] = est_hsq
        dict_summary["hsq_stderr"] = est_hsq_stderr

    # write summary
    with open(out_prefix + ".summary.json", "w") as f:
        json.dump(dict_summary, f, indent=4)
    admix.logger.info(f"Summary written to {out_prefix}.summary.json")


def meta_analyze_genet_cor(loglkl_files):
    """Meta-analyze the results of genetic correlation analysis.

    Parameters
    ----------
    loglkl_files : str
        file patterns of log-likelihood curve files
    """

    loglkl_files = glob.glob(loglkl_files)

    rg_list = None
    total_dense_loglik: np.ndarray = 0
    total_n = 0
    for f in loglkl_files:
        df_loglkl = pd.read_csv(f, sep="\t")
        if rg_list is None:
            rg_list = df_loglkl["rg"].values
            dense_rg_list = np.linspace(min(rg_list), max(rg_list), 1001)
        else:
            assert np.all(rg_list == df_loglkl["rg"].values)

        total_dense_loglik += CubicSpline(rg_list, df_loglkl["loglkl"].values)(dense_rg_list)
        # load f.replace(".loglkl.txt", ".summary.json")
        total_n += json.load(open(f.replace(".loglkl.txt", ".summary.json")))["n"]

    rg_mode = dense_rg_list[total_dense_loglik.argmax()]

    pval_rg_1 = stats.chi2.sf((total_dense_loglik.max() - total_dense_loglik[-1]) * 2, df=1)

    print(f"Meta-analysis results across {len(loglkl_files)} files")
    print("-" * 37)
    print(f"rg mode  = {rg_mode:4g}")
    for ci in [0.5, 0.95]:
        rg_hpdi = admix.data.hdi(dense_rg_list, total_dense_loglik, ci=ci)
        if isinstance(rg_hpdi, List):
            intervals = [f"[{i[0]:.4g}, {i[1]:.4g}]" for i in rg_hpdi]
            admix.logger.warning(
                f"Multiple intervals for ci={ci}, indicating some issues of model fit."
                " Please inspect log-liklihood curves for each trait."
            )
            print(f"{ci * 100:g}% HPDI = {' '.join(intervals)}")
        else:
            assert len(rg_hpdi) == 2
            print(f"{ci * 100:g}% HPDI = [{rg_hpdi[0]:.4g}, {rg_hpdi[1]:.4g}]")
    print(f"Null (rg = 1) p-value: {pval_rg_1:.4g}")
    print(f"Average N={int(np.round(total_n / len(loglkl_files)))}")


def admix_grm_rho(prefix: str, out_dir: str, rho_list=np.linspace(0, 1.0, 21)) -> None:
    """
    DEPRECATED. Will be removed in future versions.

    Build the GRM for a given rho list

    Parameters
    ----------
    prefix : str
        Prefix of the GRM files, with .K1.grm.bin and .K2.grm.bin
    out_dir : str
        folder to store the output files
    rho_list : list, optional
        List of rho values, by default np.linspace(0, 1.0, 21)
    """
    log_params("admix-grm-rho", locals())

    K1, df_id1, n_snps1 = admix.tools.gcta.read_grm(prefix + ".K1")
    K2, df_id2, n_snps2 = admix.tools.gcta.read_grm(prefix + ".K2")
    assert df_id1.equals(df_id2)
    assert np.allclose(n_snps1, n_snps2)
    df_id = df_id1
    n_snps = n_snps1

    os.makedirs(out_dir, exist_ok=True)
    for rho in tqdm(rho_list):
        K = K1 + K2 * rho

        name = os.path.join(out_dir, f"rho{int(rho * 100)}")
        admix.tools.gcta.write_grm(
            name,
            K=K,
            df_id=df_id,
            n_snps=n_snps,
        )


def estimate_genetic_cor(
    pheno: str,
    out_dir: str,
    grm_dir: str = None,
    grm_prefix: str = None,
    quantile_normalize: bool = True,
    n_thread: int = 2,
):
    """
    DEPRECATED. Will be removed in future versions.
    Estimate genetic correlation from a set of GRM files (with different rho values)

    Parameters
    ----------
    pheno : str
        phenotype file, the 1st column contains ID, 2nd column contains phenotype, and
        the rest of columns are covariates.
    out_dir : str
        folder to store the output files
    grm_dir : str
        folder containing GRM files

    quantile_normalize: bool
        whether to perform quantile normalization for both phenotype and each column of covariates
    n_thread : int, optional
        number of threads, by default 2
    """
    log_params("estimate-genetic-cor", locals())

    # either grm_dir or grm_prefix must be specified
    assert (grm_dir is not None) + (
        grm_prefix is not None
    ) == 1, "Either grm_dir or grm_prefix must be specified"

    # compile phenotype and covariates
    df_pheno = pd.read_csv(pheno, sep="\t", index_col=0)
    df_pheno.index = df_pheno.index.astype(str)

    # subset for individuals with non-nan value in df_trait
    trait_col = df_pheno.columns[0]
    covar_cols = df_pheno.columns[1:]

    # filter out individuals with missing phenotype
    df_pheno = df_pheno[df_pheno[trait_col].notna()]

    df_trait = df_pheno[[trait_col]].copy()
    df_covar = df_pheno[covar_cols].copy()
    df_covar = admix.data.convert_dummy(df_covar)
    if quantile_normalize:
        # perform quantile normalization
        for col in df_trait.columns:
            df_trait[col] = admix.data.quantile_normalize(df_trait[col])

        for col in df_covar.columns:
            df_covar[col] = admix.data.quantile_normalize(df_covar[col])

    # fill na with column mean
    df_covar.fillna(df_covar.mean(), inplace=True)

    df_id = pd.DataFrame(
        {"FID": df_trait.index.values, "IID": df_trait.index.values},
        index=df_trait.index.values,
    )
    df_trait = pd.merge(df_id, df_trait, left_index=True, right_index=True)
    df_covar = pd.merge(df_id, df_covar, left_index=True, right_index=True)

    os.makedirs(out_dir, exist_ok=True)

    if grm_dir is not None:
        # fit different rho
        grm_prefix_list = [
            p.split("/")[-1][: -len(".grm.bin")]
            for p in glob.glob(os.path.join(grm_dir, "*.grm.bin"))
        ]
    else:
        assert grm_prefix is not None
        grm_dir = os.path.dirname(grm_prefix)
        grm_prefix_list = [grm_prefix.split("/")[-1]]

    for grm_prefix in grm_prefix_list:
        grm = os.path.join(grm_dir, grm_prefix)
        out_prefix = os.path.join(out_dir, grm_prefix)
        if not os.path.exists(out_prefix + ".hsq"):
            admix.tools.gcta.reml(
                grm_path=grm,
                df_pheno=df_trait,
                df_covar=df_covar,
                out_prefix=out_prefix,
                n_thread=n_thread,
                est_fix=True,
            )



import numpy as np
from scipy import stats
import scipy
from typing import Union, List, Tuple


def hdi(
    x: np.ndarray, loglik: np.ndarray, ci: float = 0.95
) -> Union[List[Tuple], Tuple]:
    """
    Find the high density interval for 1-dimensional likelihood curve.

    Parameters
    ----------
    x : np.ndarray
        1-dimensional data
    loglik : np.ndarray
        log-likelihood of the corresponding data
    ci : float
        targeted interval

    Returns
    -------
    Tuple, List[Tuple]
        High density interval for the data. Returns a tuple denoting [low_ci, high_ci].
        Returns a list of tuples in cases where the likelihood curve is not concave (
        typically indicating the low sample size when maximizing the likelihood).
    """
    prob = np.exp(loglik - np.max(loglik))
    prob /= prob.sum()
    sorted_prob = np.sort(prob)[::-1]
    # critical value
    crit = sorted_prob[np.argmax(np.cumsum(sorted_prob) >= ci)]
    np.where(prob > crit)
    hdi_index = np.where(prob > crit)[0]

    from itertools import groupby
    from operator import itemgetter

    intervals: List[Tuple] = []
    for k, g in groupby(enumerate(hdi_index), lambda ix: ix[0] - ix[1]):
        hdi_index = list(map(itemgetter(1), g))
        intervals.append((x[hdi_index[0]], x[hdi_index[-1]]))
    if len(intervals) == 1:
        return intervals[0]
    else:
        return intervals


def lambda_gc(pval, bootstrap_ci=False, n_resamples=499):
    def _lambda(pval):
        chi2 = stats.norm.ppf(pval / 2) ** 2
        return np.quantile(chi2, 0.5) / stats.chi2.ppf(0.5, 1)

    from scipy.stats import bootstrap

    est = _lambda(pval)
    if bootstrap_ci:
        res = bootstrap(
            (pval,),
            _lambda,
            axis=-1,
            vectorized=False,
            n_resamples=n_resamples,
        )
        ci = res.confidence_interval
        return est, (ci[0], ci[1])
    else:
        return est


def pval2chisq(pval: np.ndarray, two_sided: bool = True):
    """Convert p-value to chisq

    Parameters
    ----------
    pval : np.ndarray
        p-value
    two_sided : bool
        Whether to use two-sided test

    Returns
    -------
    np.ndarray
        z-score
    """
    pval = np.array(pval)
    if two_sided:
        return stats.norm.ppf(pval / 2) ** 2
    else:
        return stats.norm.ppf(pval) ** 2


def pval2zsc(pval):
    return -scipy.stats.norm.ppf(pval)


def zsc2pval(zscore: np.ndarray, two_sided: bool = True):
    """Convert z-score to p-value

    Parameters
    ----------
    zscore : np.ndarray
        z-score
    two_sided : bool
        Whether to use two-sided test

    Returns
    -------
    np.ndarray
        p-value
    """
    if two_sided:
        return 2 * (1 - stats.norm.cdf(np.abs(zscore)))
    else:
        return 1 - stats.norm.cdf(zscore)


def quantile_normalize(val):
    from scipy.stats import rankdata, norm

    val = np.array(val)
    non_nan_index = ~np.isnan(val)
    results = np.full(val.shape, np.nan)
    results[non_nan_index] = norm.ppf(
        (rankdata(val[non_nan_index]) - 0.5) / len(val[non_nan_index])
    )
    return results


def quad_form(x, A):
    return np.dot(np.dot(x.T, A), x)


def chi2_to_logpval(chi2, dof=1):
    return scipy.stats.chi2.logsf(chi2, dof)


def deming_regression(
    x: np.ndarray,
    y: np.ndarray,
    sx: np.ndarray = None,
    sy: np.ndarray = None,
    no_intercept: bool = False,
):
    """Deming regression.

    Parameters
    ----------
    x : np.ndarray
        x variables
    y : np.ndarray
        y variables
    sx : np.ndarray, optional
        standard errors of x variables, by default None
    sy : np.ndarray, optional
        standard errors of y variables, by default None
    no_intercept : bool, optional
        whether not to fit intercept or not, by default False

    Returns
    -------
    If no_intercept is False:
        no intercept is fit, return a single slope
    If no_intercept is True:
        intercept is fit, return slope and intercept
    """

    def no_intercept_func(B, x):
        return B[0] * x

    if no_intercept:
        model = scipy.odr.Model(no_intercept_func)
        odr = scipy.odr.ODR(scipy.odr.RealData(x, y, sx=sx, sy=sy), model, beta0=[1])
        fit = odr.run()
        return fit.beta[0]
    else:
        model = scipy.odr.unilinear
        odr = scipy.odr.ODR(scipy.odr.RealData(x, y, sx=sx, sy=sy), model)
        fit = odr.run()
        return fit.beta[0], fit.beta[1]


def meta_analysis(
    effects: np.ndarray, se: np.ndarray, method="random", weights: np.ndarray = None
) -> float:
    """Meta analysis of effects

    Parameters
    ----------
    effects : np.ndarray
        effects array
    se : np.ndarray
        effects array
    method : str, optional
        method for meta-analysis, by default "random"
    weights : np.ndarray, optional
        weight for different effects, by default None

    Returns
    -------
    float
        single number summarizing the meta-analysis results
    """
    # From Omer Weissbrod
    assert method in ["fixed", "random"]
    d = effects
    variances = se ** 2

    # compute random-effects variance tau2
    vwts = 1.0 / variances
    fixedsumm = vwts.dot(d) / vwts.sum()
    Q = np.sum(((d - fixedsumm) ** 2) / variances)
    df = len(d) - 1
    tau2 = np.maximum(0, (Q - df) / (vwts.sum() - vwts.dot(vwts) / vwts.sum()))

    # defing weights
    if weights is None:
        if method == "fixed":
            wt = 1.0 / variances
        else:
            wt = 1.0 / (variances + tau2)
    else:
        wt = weights

    # compute summtest
    summ = wt.dot(d) / wt.sum()
    if method == "fixed":
        varsum = np.sum(wt * wt * variances) / (np.sum(wt) ** 2)
    else:
        varsum = np.sum(wt * wt * (variances + tau2)) / (np.sum(wt) ** 2)

    summary = summ
    se_summary = np.sqrt(varsum)

    return summary, se_summary


"""
External tools
"""

from dask.array.core import Array
import dask.array as da
from ._utils import get_dependency, get_cache_data
from . import gcta, plink2, liftover, plink
from ._standalone import hapgen2, admix_simu, haptools_simu_admix


__all__ = [
    "plink2",
    "gcta",
    "lift_over",
    "plink_assoc",
    "plink_read_bim",
    "plink_read_fam",
]



import statsmodels.api as sm
import numpy as np
import pandas as pd
from scipy import stats
from tqdm import tqdm
import admix
from typing import Any, Dict, List, Optional
import dask.array as da
import admix
from statsmodels.tools import sm_exceptions
import warnings
import itertools

warnings.filterwarnings(action="error", category=sm_exceptions.ValueWarning)

__all__ = ["marginal"]


def _format_block_test(
    var: np.ndarray,
    cov: np.ndarray,
    pheno: np.ndarray,
    var_size: int,
    test_vars: np.ndarray,
):
    n_indiv = var.shape[0]
    assert (
        cov.shape[0] == n_indiv
    ), "Number of individuals in genotype and covariate do not match"
    assert pheno.ndim == 1, "Phenotype must be a vector"
    assert (
        pheno.shape[0] == n_indiv
    ), "Number of individuals in genotype and phenotype do not match"
    assert var_size > 0, "Variable size must be greater than 0"
    assert (
        var.shape[1] % var_size == 0
    ), f"Number of variables in var ({var.shape[1]}) must be a multiple of var_size ({var_size})"
    n_var = var.shape[1] // var_size

    if isinstance(test_vars, List):
        test_vars = np.array(test_vars)

    assert np.all(test_vars < var_size), "test_vars must be less than var_size"

    # fill covariates
    n_cov = cov.shape[1]

    if isinstance(var, da.Array):
        var = var.compute()

    assert isinstance(var, np.ndarray), "var must be a numpy array"
    return var, n_var, n_cov, test_vars


def _block_het_test(
    var: np.ndarray,
    cov: np.ndarray,
    pheno: np.ndarray,
    var_size: int,
    test_vars: np.ndarray,
    fast: bool,
    family: str,
    logistic_kwargs: Dict[str, Any] = dict(),
):
    admix.logger.info(
        "Currently HET test is implemented through statsmodels, which can be slow. "
        "Pass in small amount of data whenever possible."
    )

    ## Format data
    var, n_var, n_cov, test_vars = _format_block_test(
        var, cov, pheno, var_size, test_vars
    )

    # format reduced variable matrix
    n_indiv = var.shape[0]
    reduced_var_size = var_size - len(test_vars) + 1
    design_full = np.zeros((n_indiv, var_size + n_cov))
    design_full[:, var_size : var_size + n_cov] = cov
    design_reduced = np.zeros((n_indiv, reduced_var_size + n_cov))
    design_reduced[:, reduced_var_size : reduced_var_size + n_cov] = cov
    other_vars = np.array([i for i in range(var_size) if i not in test_vars])
    shared_param_index = np.concatenate(
        [other_vars, np.arange(var_size, var_size + n_cov)]
    ).astype(int)

    # statsmodels implementation
    if family == "linear":
        reg_method = lambda pheno, design, start_params=None: sm.OLS(
            pheno, design, missing="drop"
        ).fit(disp=0, start_params=start_params)

    elif family == "logistic":
        reg_method = lambda pheno, design, start_params=None: sm.Logit(
            pheno, design, missing="drop"
        ).fit(disp=0, start_params=start_params)
    else:
        raise NotImplementedError

    # legacy implementation of using F-test
    # ftest_mat = np.zeros([len(test_vars) - 1, design_full.shape[1]])
    # for i in range(len(test_vars) - 1):
    #     ftest_mat[i, test_vars[i]] = 1
    #     ftest_mat[i, test_vars[i + 1]] = -1

    res = np.zeros((n_var, var_size * 2 + 2))

    for i in range(n_var):
        design_full[:, 0:var_size] = var[:, i * var_size : (i + 1) * var_size]
        model_full = reg_method(pheno, design_full)

        # coefficients
        res[i, 0 : var_size * 2 : 2] = model_full.params[0:var_size]
        # standard errors
        res[i, 1 : var_size * 2 : 2] = model_full.bse[0:var_size]
        res[i, -2] = model_full.nobs

        design_reduced[:, 0] = var[:, test_vars + i * var_size].sum(axis=1)
        if len(other_vars) > 0:
            design_reduced[:, 1:reduced_var_size] = var[:, other_vars + i * var_size]

        model_reduced = reg_method(
            pheno,
            design_reduced,
            start_params=np.concatenate([[0.0], model_full.params[shared_param_index]]),
        )

        if family == "linear" or family == "logistic":
            p = stats.chi2.sf(
                -2 * (model_reduced.llf - model_full.llf),
                (model_full.df_model - model_reduced.df_model),
            )
            res[i, -1] = p
        else:
            raise NotImplementedError
    return res


def _block_test(
    var: np.ndarray,
    cov: np.ndarray,
    pheno: np.ndarray,
    var_size: int,
    test_vars: np.ndarray,
    fast: bool,
    family: str,
    logistic_kwargs: Dict[str, Any] = dict(),
) -> np.ndarray:
    """
    Perform association testing for a block of variables

    Parameters
    ----------
    var : np.ndarray
        (n_indiv, n_var x var_size) variable matrix
    cov : np.ndarray
        (n_indiv, n_cov) covariate matrix
    pheno : np.ndarray
        (n_snp) phenotype matrix
    var_size : int
        Number of variables for each test
    test_vars : List[int]
        Index of variables to test

    Returns
    -------
    np.ndarray
        (n_snp, n_test_var * 2 + 2) association testing information.
        - The first n_test_var * 2 columns are the effect size (odd columns)
        and standard error (even columns) for each variable.
        - The last two columns are the number of individuals and p-value.
    """
    var, n_var, n_cov, test_vars = _format_block_test(
        var, cov, pheno, var_size, test_vars
    )

    if fast:
        try:
            import tinygwas
        except ImportError:
            raise ImportError(
                "\nplease install tinygwas:\n\n"
                "\tpip install git+https://github.com/bogdanlab/tinygwas.git#egg=tinygwas"
            )

        res = np.zeros((n_var, var_size * 2 + 2))
        if family == "linear":
            tinygwas.linear_f_test(var, cov, pheno, var_size, test_vars, res)
            f_stats = res[:, -1]
            n_indiv_test = res[:, -2]

            res[:, -1] = stats.f.sf(
                f_stats, len(test_vars), n_indiv_test - n_cov - var_size
            )

        elif family == "logistic":
            if "max_iter" not in logistic_kwargs:
                logistic_kwargs["max_iter"] = 100
            if "tol" not in logistic_kwargs:
                logistic_kwargs["tol"] = 1e-6

            tinygwas.logistic_lrt(
                var,
                cov,
                pheno,
                var_size,
                test_vars,
                res,
                logistic_kwargs["max_iter"],
                logistic_kwargs["tol"],
            )
            # convert lrt diff to pvalue
            res[:, -1] = stats.chi2.sf(2 * res[:, -1], len(test_vars))
        else:
            raise ValueError(f"Unknown family: {family}")
    else:
        # statsmodels implementation
        n_indiv = var.shape[0]
        design = np.zeros((n_indiv, var_size + n_cov))
        design[:, var_size : var_size + n_cov] = cov

        if family == "linear":
            reg_method = lambda pheno, design, start_params=None: sm.OLS(
                pheno, design, missing="drop"
            ).fit(disp=0, start_params=start_params)
        elif family == "logistic":
            reg_method = lambda pheno, design, start_params=None: sm.Logit(
                pheno, design, missing="drop"
            ).fit(disp=0, start_params=start_params)
        else:
            raise NotImplementedError

        reduced_index = np.concatenate(
            [
                [i for i in range(var_size) if i not in test_vars],
                np.arange(var_size, var_size + n_cov),
            ]
        ).astype(int)

        f_test_r_matrix = np.zeros((len(test_vars), design.shape[1]))
        for i, v in enumerate(test_vars):
            f_test_r_matrix[i, v] = 1

        res = np.zeros((n_var, var_size * 2 + 2))

        for i in range(n_var):
            design[:, 0:var_size] = var[:, i * var_size : (i + 1) * var_size]
            model = reg_method(pheno, design)
            # coefficients
            res[i, 0 : var_size * 2 : 2] = model.params[0:var_size]
            # standard errors
            res[i, 1 : var_size * 2 : 2] = model.bse[0:var_size]
            res[i, -2] = model.nobs
            # sample size
            if family == "linear":
                # f-test using statsmodels
                try:
                    p = model.f_test(f_test_r_matrix).pvalue.item()
                except sm_exceptions.ValueWarning:
                    p = np.nan
                res[i, -1] = p

            elif family == "logistic":
                # more than one test variables
                model_reduced = reg_method(
                    pheno,
                    design[:, reduced_index],
                    start_params=model.params[reduced_index],
                )
                # determine p-values using difference in log-likelihood
                # and difference in degrees of freedom
                p = stats.chi2.sf(
                    -2 * (model_reduced.llf - model.llf),
                    (model.df_model - model_reduced.df_model),
                )
                res[i, -1] = p
    return res


def marginal(
    dset: admix.Dataset = None,
    geno: da.Array = None,
    lanc: da.Array = None,
    pheno: np.ndarray = None,
    n_anc: Optional[int] = None,
    cov: Optional[np.ndarray] = None,
    method: str = "ATT",
    family: str = "linear",
    fast: bool = True,
):
    """Marginal association testing for one SNP at a time

    Parameters
    ----------
    dset : admix.Dataset
        data set
    geno : da.Array
        genotype (n_snp, n_indiv, 2) matrix
    lanc : da.Array
        local ancestry (n_snp, n_indiv, 2) matrix
    pheno : np.ndarray
        phenotype (n_snp, )
    n_anc : int
        number of ancestral populations, if not specified, inferred from lanc
    cov : np.ndarray, optional
        Covariate matrix, by default None. Do NOT include `1` intercept
    method : str, optional
        methods used for association testing, by default "ATT", one of
        ["ATT", "TRACTOR", "JOINT", "ADM", "SNP1", "ASE", "HET"]
    family : str, optional
        family of phenotype, by default "linear"
    fast : bool, optional
        use fast implementation, by default True

    Returns
    -------
    np.ndarray
        Association p-values for each SNP being tested

    """

    assert family in ["linear", "logistic"], "Unknown family"

    # check phenotype
    # nan values are not allowed
    assert pheno is not None, "Must specify `pheno`"
    assert np.all(np.isfinite(pheno)), "pheno must not contain NaN values"
    if family == "logistic":
        assert np.all(
            (pheno == 0) | (pheno == 1)
        ), "When family='logistic', pheno must be 0 or 1"

    if family == "logistic":
        if len(pheno) < 1000:
            admix.logger.warn(
                "logistic family is known to be unstable with small sample size (N < 1,000)"
                + "NaN values in the returned p-values may be caused by this."
            )

    # format data
    assert method in ["ATT", "TRACTOR", "JOINT", "ADM", "SNP1", "ASE", "HET"]
    if dset is not None:
        assert (geno is None) and (
            lanc is None
        ), "Cannot specify both `dset` and `geno`, `lanc`"
        geno = dset.geno
        lanc = dset.lanc
        n_anc = dset.n_anc
    else:
        assert (
            (geno is not None) and (lanc is not None) and (n_anc is not None)
        ), "Must specify `dset` or (`geno`, `lanc`, `n_anc`)"
        # convert geno and lanc to da.Array when necessary
        if not isinstance(geno, da.Array):
            geno = da.from_array(geno, chunks=-1)
        if not isinstance(lanc, da.Array):
            lanc = da.from_array(lanc, chunks=-1)

    assert np.all(geno.shape == lanc.shape), "geno and lanc must have same shape"
    n_snp, n_indiv = geno.shape[0:2]

    # process covariates
    if cov is None:
        cov = np.ones((n_indiv, 1))
    else:
        assert cov.shape[0] == n_indiv, "cov must have same number of rows as pheno"
        # prepend a column of ones to the covariates
        cov = np.hstack((np.ones((n_indiv, 1)), cov))

    assert cov is not None

    # impute missing values when needed
    if np.isnan(cov).any():
        admix.logger.info(
            "NaN found in covariates, impute with column mean for each covariate."
        )
        # fill nan with column mean
        debug_old_mean = np.nanmean(cov, axis=0)
        cov = admix.data.impute_with_mean(cov, axis=0)
        assert np.allclose(
            np.nanmean(cov, axis=0), debug_old_mean, equal_nan=True
        ), "NaN imputation failed"

    # check covariates must be full rank
    assert np.linalg.matrix_rank(cov) == cov.shape[1], "Covariates must be of full rank"

    if method == "ATT":
        # test genotype dosage
        var = geno.sum(axis=2).swapaxes(0, 1)
        var_size = 1
        var_names = ["G"]
        test_vars = [0]

    elif method == "SNP1":
        # test genotype dosage, condition on local ancestry
        var = da.empty((n_indiv, n_snp * n_anc))
        var[:, 0::n_anc] = geno.sum(axis=2).swapaxes(0, 1)
        for i in range(n_anc - 1):
            var[:, (1 + i) :: n_anc] = (lanc == i).sum(axis=2).swapaxes(0, 1)
        var_size = n_anc
        var_names = ["G"] + [f"L{i + 1}" for i in range(n_anc - 1)]
        test_vars = [0]

    elif method == "TRACTOR":
        # test ancestry-specfic genotype dosage, condition on local ancestry
        allele_per_anc = admix.data.allele_per_anc(
            geno,
            lanc,
            n_anc=n_anc,
        ).swapaxes(0, 1)
        var = da.empty((n_indiv, n_snp * (2 * n_anc - 1)))

        # allele per ancestor per SNP
        for i in range(n_anc):
            var[:, i :: (2 * n_anc - 1)] = allele_per_anc[:, :, i]

        # number of ancestries per SNP
        for i in range(n_anc - 1):
            var[:, (i + n_anc) :: (2 * n_anc - 1)] = (
                (lanc == i).sum(axis=2).swapaxes(0, 1)
            )
        var_size = 2 * n_anc - 1
        var_names = [f"G{i + 1}" for i in range(n_anc)] + [
            f"L{i + 1}" for i in range(n_anc - 1)
        ]
        test_vars = [i for i in range(n_anc)]

    elif method == "JOINT":
        # joint test of ancestry-specfic genotype dosage AND local ancestry
        allele_per_anc = admix.data.allele_per_anc(
            geno,
            lanc,
            n_anc=n_anc,
        ).swapaxes(0, 1)
        var = da.empty((n_indiv, n_snp * (2 * n_anc - 1)))

        # allele per ancestor per SNP
        for i in range(n_anc):
            var[:, i :: (2 * n_anc - 1)] = allele_per_anc[:, :, i]

        # number of ancestries per SNP
        for i in range(n_anc - 1):
            var[:, (i + n_anc) :: (2 * n_anc - 1)] = (
                (lanc == i).sum(axis=2).swapaxes(0, 1)
            )
        var_size = 2 * n_anc - 1
        var_names = [f"G{i + 1}" for i in range(n_anc)] + [
            f"L{i + 1}" for i in range(n_anc - 1)
        ]
        # test both genotype and local ancestry
        test_vars = [i for i in range(var_size)]

    elif method == "ASE":
        # test ancestry-specfic genotype dosage, without conditioning on local ancestry
        allele_per_anc = admix.data.allele_per_anc(geno, lanc, n_anc=n_anc).swapaxes(
            0, 1
        )
        var = da.empty((n_indiv, n_snp * n_anc))

        for i in range(n_anc):
            var[:, i::n_anc] = allele_per_anc[:, :, i]

        var_size = n_anc
        var_names = [f"G{i + 1}" for i in range(n_anc)]
        test_vars = [i for i in range(n_anc)]

    elif method == "ADM":
        # test local ancestry
        var = da.empty((n_indiv, n_snp * (n_anc - 1)))
        for i in range(n_anc - 1):
            var[:, i :: (n_anc - 1)] = (lanc == i).sum(axis=2).swapaxes(0, 1)
        var_size = n_anc - 1
        var_names = [f"L{i + 1}" for i in range(n_anc - 1)]
        test_vars = [i for i in range(n_anc - 1)]

    elif method == "HET":
        allele_per_anc = admix.data.allele_per_anc(geno, lanc, n_anc=n_anc).swapaxes(
            0, 1
        )
        var = da.empty((n_indiv, n_snp * n_anc))

        for i in range(n_anc):
            var[:, i::n_anc] = allele_per_anc[:, :, i]

        var_size = n_anc
        var_names = [f"G{i + 1}" for i in range(n_anc)]
        test_vars = [i for i in range(n_anc)]

    else:
        raise NotImplementedError

    # iterate over block of SNPs
    assert var.shape[1] % var_size == 0, "var must have multiple of `var_size` columns"
    assert var.shape[1] / var_size == n_snp

    res = []
    # block-by-block computation based on the chunk size of the `geno` array
    if geno is not None:
        snp_chunks = geno.chunks[0]
    else:
        assert lanc is not None
        snp_chunks = lanc.chunks[0]

    for snp_start, snp_stop in tqdm(
        admix.data.index_over_chunks(snp_chunks),
        desc="admix.assoc.marginal",
        total=len(snp_chunks),
    ):
        if method == "HET":
            test_func = _block_het_test
        else:
            test_func = _block_test
        # test each SNP in block
        res.append(
            test_func(
                var=var[:, snp_start * var_size : snp_stop * var_size],
                cov=cov,
                pheno=pheno,
                var_size=var_size,
                test_vars=test_vars,
                family=family,
                fast=fast,
            )
        )
    # columns will be BETA1, SE1, BETA2, SE2, ... with n_anc
    columns = list(
        itertools.chain.from_iterable([[f"{v}_BETA", f"{v}_SE"] for v in var_names])
    ) + ["N", "P"]

    df_res = pd.DataFrame(
        np.concatenate(res),
        columns=columns,
        index=dset.snp.index if dset is not None else None,
    ).astype({"N": "int"})
    df_res.loc[df_res.P.isna(), df_res.columns != "N"] = np.nan

    return df_res


def marginal_simple(dset: admix.Dataset, pheno: np.ndarray) -> np.ndarray:
    """Simple marginal association testing for one SNP at a time

    Useful in simulation study because this will be very fast

    Parameters
    ----------
    dset : admix.Dataset
        Dataset containing the (n_indiv, n_snp) genotype matrix, dset.geno
    pheno : np.ndarray
        (n_snp, n_sim) phenotype matrix

    Returns
    -------
    coef : np.ndarray
        (n_snp, n_sim) marginal association coefficient
    coef_se: np.ndarray
        (n_snp, n_sim) marginal association coefficient standard error
    zscores : np.ndarray
        (n_snp, n_sim) association z-scores for each SNP being tested

    Examples
    --------

    To check the consistency of results of standard methods

    >>> n_indiv = dset_admix.dims["indiv"]
    >>> n_cov = 1

    >>> geno = _impute_with_mean(dset_admix.geno.values)
    >>> geno = (geno - geno.mean(axis=0)) / geno.std(axis=0)

    >>> f_stats = tinygwas.linear_f_test(geno, np.ones((n_indiv, 1)), sim["pheno"][:, 0], 1, [0])
    >>> p_vals = stats.f.sf(f_stats, 1, n_indiv - n_cov - 1)
    >>> zscores2 = stats.norm.ppf(p_vals / 2) * np.sign(zscores[:, 0])

    >>> dset = admix.Dataset({"geno": (["indiv", "snp"], geno), "pheno": (["snp", "sim"], pheno)})
    >>> zscores = marginal_simple(dset, pheno)

    """
    geno = dset.geno
    n_indiv, n_snp = geno.shape
    assert (
        n_indiv == pheno.shape[0]
    ), "Number of individuals in genotype and phenotype do not match"
    n_sim = pheno.shape[1]

    # center phenotype for each simulation
    Y = pheno - pheno.mean(axis=0)
    X = geno - da.nanmean(geno, axis=0)

    XtY, snp_var = admix.data.geno_mult_mat(
        X, Y, transpose_geno=True, return_snp_var=True
    )
    XtX = snp_var * n_indiv

    coef = XtY / XtX[:, np.newaxis]
    coef_var = np.var(Y, axis=0) / XtX[:, np.newaxis]
    coef_se = np.sqrt(coef_var)

    zscores = coef / coef_se

    return coef, coef_se, zscores


# def mixscore_wrapper(pheno, anc, geno, theta,
#                     scores=["ADM", "ATT", "MIX", "SNP1", "SUM"],
#                     mixscore_path="/u/project/pasaniuc/kangchen/tractor/software/mixscore-1.3/bin/mixscore",
#                     verbose=False):
#     """
#     A python wrapper for mixscore
#
#     Args
#     ----
#     pheno: phenotypes
#     anc: ancestry
#     geno: genotype
#     theta: global ancestry component
#     """
#
#     tmp = tempfile.TemporaryDirectory()
#     tmp_dir = tmp.name
#
#     n_sample = len(pheno)
#     n_snp = anc.shape[1]
#
#     write_int_mat(join(tmp_dir, "pheno"), pheno.reshape((1, -1)))
#     write_int_mat(join(tmp_dir, "anc"), anc.T)
#     write_int_mat(join(tmp_dir, "geno"), geno.T)
#     np.savetxt(join(tmp_dir, "theta"), theta, fmt='%.6f')
#
#     param = {"nsamples": str(n_sample),
#               "nsnps": str(n_snp),
#               "phenofile": join(tmp_dir, "pheno"),
#               "ancfile": join(tmp_dir, "anc"),
#               "genofile": join(tmp_dir, "geno"),
#               "thetafile": join(tmp_dir, "theta"),
#               "outfile": join(tmp_dir, "out")}
#
#     with open(join(tmp_dir, "param"), 'w') as f:
#         f.writelines([k + ':' + param[k] + '\n' for k in param])
#
#     rls_dict = {}
#     for name in scores:
#         if verbose:
#             print(f"Calculating {name}...")
#
#         cmd = ' '.join([mixscore_path, name, f"{tmp_dir}/param"])
#         subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
#         with open(param["outfile"]) as f:
#             out = [line.strip() for line in f.readlines()]
#         rls_dict[name] = out
#     tmp.cleanup()
#     score_df = pd.DataFrame(rls_dict).apply(pd.to_numeric, errors='coerce')
#     # convert to p-value
#     for name in score_df.columns:
#         score_df[name] = stats.chi2.sf(score_df[name], df=(2 if name == "SUM" else 1))
#     return score_df



from typing import List
from ._utils import get_dependency
import subprocess
import numpy as np
from typing import Union
import pandas as pd
import admix
import os
import glob


def run(cmd: str, **kwargs):
    """Shortcut for running plink commands

    Parameters
    ----------
    cmd : str
        plink command
    """
    bin_path = get_dependency("plink")
    add_cmds = [f" --{k.replace('_', '-')} {kwargs[k]}" for k in kwargs]
    cmd += " ".join(add_cmds)

    subprocess.check_call(f"{bin_path} {cmd}", shell=True)


def merge(bfiles: List[str], out: str, **kwargs):
    """Shortcut for merging a list of bfiles

    Parameters
    ----------
    bfiles : List[str]
        List of bfile paths
    out : str
        out prefix
    """

    assert len(bfiles) >= 2
    merge_list = out + ".merge_list"
    np.savetxt(merge_list, bfiles[1:], fmt="%s", delimiter="\n")
    cmd = f"--bfile {bfiles[0]} --merge-list {merge_list} --keep-allele-order --make-bed --out {out}"
    run(cmd)


def read_bim(bim: Union[str, List[str]]):
    if isinstance(bim, list):
        return pd.concat(
            [
                pd.read_csv(
                    b,
                    header=None,
                    delim_whitespace=True,
                    names=["CHR", "SNP", "CM", "POS", "A1", "A2"],
                )
                for b in bim
            ]
        ).reset_index(drop=True)
    elif isinstance(bim, str):
        return pd.read_csv(
            bim,
            header=None,
            delim_whitespace=True,
            names=["CHR", "SNP", "CM", "POS", "A1", "A2"],
        )
    else:
        raise ValueError("`bim` should either be a string or a list of string")


def read_fam(fam: str):
    return pd.read_csv(
        fam,
        header=None,
        delim_whitespace=True,
        usecols=[0, 1],
        names=["FID", "IID"],
    ).astype(str)


def clump(
    bfile: str,
    assoc_path: str,
    out_prefix: str,
    p1: float = 5e-8,
    p2: float = 1e-4,
    r2: float = 0.1,
    kb=3000,
    **kwargs,
):
    """
    Wrapper for plink2 clump

    Parameters
    ----------
    bfile : str
        Path to plink binary file
    assoc_path : str
        Path to association file, assumed to be PLINK2 format, the function will
        perform the necessary conversion
    out_prefix : str
        Prefix for output files
    p1 : float, optional
        p-value threshold for clumping, by default 5e-8
    p2 : float, optional
        p-value threshold for clumping, by default 1e-4
    r2 : float, optional
        r2 threshold for clumping, by default 0.1
    kb : int, optional
        Number of kb to use for clumping, by default 3000
    """
    tmp_prefix = out_prefix + ".admix_plink_clump_tmp"
    # convert plink2 association to plink1 format ID -> SNP
    import shutil

    from_file = open(assoc_path)
    to_file = open(tmp_prefix + ".assoc", "w")
    to_file.writelines(from_file.readline().replace("ID", "SNP"))
    shutil.copyfileobj(from_file, to_file)
    from_file.close()
    to_file.close()
    cmds = [
        f"--bfile {bfile} --clump {tmp_prefix + '.assoc'}",
        f"--clump-p1 {p1} --clump-p2 {p2} --clump-r2 {r2} --clump-kb {kb}",
        f"--out {tmp_prefix}",
    ]

    admix.tools.plink.run(" ".join(cmds), **kwargs)
    if os.path.exists(tmp_prefix + ".clumped"):
        os.rename(tmp_prefix + ".clumped", out_prefix + ".clumped")
    else:
        # no clumped region
        # write a comment to the output file
        with open(out_prefix + ".clumped", "w") as file:
            file.write("# No clumped region")

    for f in glob.glob(tmp_prefix + "*"):
        os.remove(f)



import admix
import numpy as np
from typing import List
import dapgen
import pandas as pd
import glob
import os
from ._utils import log_params
from typing import List


def prune(pfile: str, out: str, indep_pairwise_params: List = None):
    """Prune a pfile based on indep_pairwise_params

    Parameters
    ----------
    pfile : str
        pfile
    out : str
        out_prefix
    indep_pairwise_params : [type], optional
        if None, use the default [100 5 0.1]

    Returns
    -------
    out.[pgen|pvar|psam] will be created
    """
    log_params("prune", locals())

    if indep_pairwise_params is None:
        indep_pairwise_params = [100, 5, 0.1]

    admix.tools.plink2.prune(
        pfile=pfile,
        out_prefix=out,
        indep_pairwise_params=indep_pairwise_params,
    )


def pca(pfile: str, out: str, approx=False):
    """
    Perform PCA on a pgen file

    Parameters
    ----------
    pfile : str
        Path to the pgen file
    out : str
        Path to the output file
    approx : bool, optional
        If True, use the approximate algorithm.
    """
    log_params("pca", locals())

    admix.tools.plink2.pca(pfile=pfile, out_prefix=out, approx=approx)


def liftover(pfile: str, out: str, chain="hg19->hg38"):
    """
    Lift over a pgen file

    Parameters
    ----------
    pfile : str
        Path to the pgen file
    out : str
        Path to the output file
    chain : str, optional
        Chain file to use.
    """
    log_params("liftover", locals())

    admix.tools.plink2.lift_over(pfile=pfile, out_prefix=out, chain=chain)


def pfile_merge_indiv(pfile1: str, pfile2: str, out: str):
    """Merge individuals from 2 pfiles into a single pfile

    Parameters
    ----------
    pfile1 : str
        first pfile
    pfile2 : str
        second pfile
    out : str
        output pfile prefix
    """

    log_params("pfile-merge-indiv", locals())
    admix.tools.plink2.merge_indiv(pfile1=pfile1, pfile2=pfile2, out_prefix=out)
    admix.logger.info(
        f"Currently PLINK2 pmerge does not fully support merging pfiles with different individuals, writing PLINK1 bed file instead."
    )
    admix.logger.info(f"{out}.bed, {out}.bim, {out}.fam are created")


def pfile_align_snp(pfile1: str, pfile2: str, out: str):
    """align the SNP for 2 pfiles

    Parameters
    ----------
    pfile1 : str
        plink file 1
    pfile2 : str
        plink file 2
    out : str
        output prefix
    """
    log_params("pfile-align-snp", locals())
    admix.tools.plink2.align_snp(pfile1=pfile1, pfile2=pfile2, out_prefix=out)


def pfile_freq_within(
    pfile: str,
    group_col: str,
    out: str,
    groups: str = None,
):
    """calculate the within-cluster frequencies for a pfile using
    plink --freq --within

    Parameters
    ----------
    pfile : str
        plink file
    col : str
        column
    out : str
        output prefix
    groups : str, optional
        unique groups
    """
    log_params("pfile-freq-within", locals())

    # 1. convert pfile into bfile
    bfile = f"{out}-tmp"
    admix.tools.plink2.run(f"--pfile {pfile} --make-bed --out {bfile}")

    # 2. create --within file
    # https://www.cog-genomics.org/plink/1.9/input#within
    within_file = f"{out}-tmp.within"
    df_psam = dapgen.read_psam(pfile + ".psam")
    assert (
        group_col in df_psam.columns
    ), f"group-col={group_col} not in {pfile}.psam: {df_psam.columns}"
    if groups is None:
        groups = df_psam[group_col].unique()
    df_within = df_psam[df_psam[group_col].isin(groups)].copy()

    # when converting PLINK2 to PLINK1, all #FID is converted to 0
    df_within = pd.DataFrame(
        {
            "FID": ["0"] * len(df_within),
            "IID": df_within.index,
            group_col: df_within[group_col].values,
        }
    )
    df_within.to_csv(within_file, sep="\t", index=False, header=False)

    print(df_within)
    # 3. run plink --freq --within
    admix.tools.plink.run(f"--bfile {bfile} --freq --within {within_file} --out {out}")

    # 4. clean up
    for f in glob.glob(f"{out}-tmp*"):
        os.remove(f)


def subset_hapmap3(
    pfile: str, build: str, chrom: int = None, out_pfile: str = None, out: str = None
):
    log_params("subset-hapmap3", locals())
    assert (out_pfile is None) + (
        out is None
    ) == 1, "only one of out_prefix and out can be specified"

    if out_pfile is not None:
        admix.tools.plink2.subset_hapmap3(
            pfile, build=build, chrom=chrom, out_prefix=out_pfile
        )

    if out is not None:
        snp_list = admix.tools.plink2.subset_hapmap3(pfile, chrom=chrom, build=build)
        np.savetxt(out, snp_list, fmt="%s", delimiter="\n")


def subset_pop_indiv(
    pfile: str,
    out: str,
    superpop: str = None,
    exclude_pop: List[str] = None,
    pop: List[str] = None,
):
    log_params("subset-pop-indiv", locals())
    df_psam = dapgen.read_psam(pfile + ".psam")

    assert (superpop is None) + (
        pop is None
    ) == 1, "only one of superpop and pop can be specified"

    if superpop is not None:
        assert pop is None
        mask = df_psam["SuperPop"] == superpop
        if exclude_pop is not None:
            if isinstance(exclude_pop, str):
                exclude_pop = (exclude_pop,)
            assert isinstance(exclude_pop, tuple)

            mask &= ~df_psam["SuperPop"].isin(exclude_pop)

    if pop is not None:
        assert (superpop is None) and (exclude_pop is None)
        if isinstance(pop, str):
            pop = (pop,)
        assert isinstance(pop, tuple)
        mask = df_psam["Population"].isin(pop)

    indiv = df_psam.loc[mask, :].index.values
    admix.logger.info(f"{len(indiv)}/{len(df_psam)} individuals are retained")

    np.savetxt(out, indiv, fmt="%s", delimiter="\n")


def hapgen2(pfile: str, n_indiv: int, out: str, build: str, chrom: int = None):
    """Run HAPGEN2 to expand population using a PLINK file."""
    log_params("hapgen2", locals())
    admix.tools.hapgen2(
        pfile=pfile,
        chrom=chrom,
        n_indiv=n_indiv,
        out_prefix=out,
        genetic_map=build,
    )


def admix_simu(
    pfile_list: List[str],
    admix_prop: List[float],
    n_gen: int,
    n_indiv: int,
    build: str,
    out: str,
):
    """Run admix-simu to expand population using a PLINK file.

    Parameters
    ----------
    pfile_list : List[str]
        list of pgen files, with or without .pgen extension is file
    admix_prop : List[float]
        list of admix proportions
    n_gen : int
        number of generations
    n_indiv : int
        number of individuals
    build : str
        genetic map build, e.g. hg38, hg19
    out : str
        output prefix

    """
    log_params("admix-simu", locals())
    assert isinstance(pfile_list, list)
    assert isinstance(admix_prop, list)
    admix_prop = [float(p) for p in admix_prop]
    # remove .pgen extention name if present
    pfile_list = [
        os.path.splitext(p)[0] if p.endswith(".pgen") else p for p in pfile_list
    ]
    admix.logger.info(f"Received pfile_list={','.join(pfile_list)}")
    admix.tools.admix_simu(
        pfile_list=pfile_list,
        admix_prop=admix_prop,
        n_gen=n_gen,
        n_indiv=n_indiv,
        build=build,
        out_prefix=out,
    )


def haptools_simu_admix(
    pfile: str,
    admix_prop: List[float],
    pop_col: str,
    mapdir: str,
    n_gen: int,
    n_indiv: int,
    out: str,
):
    """Simulate admixture

    admix haptools-simu-admix \
        --pfile data/hm3_chrom22 \
        --admix-prop '{"FIN": 0.5, "IBS": 0.3, "JPT": 0.2}' \
        --pop-col Population \
        --mapdir data/1kg-ref-hg38/metadata/genetic_map/ \
        --n-gen 10 \
        --n-indiv 100 \
        --out test

    Parameters
    ----------
    pfile_list : List[str]
        list of pgen files, with or without .pgen extension is file
    admix_prop : List[float]
        list of admix proportions
    n_gen : int
        number of generations
    n_indiv : int
        number of individuals
    build : str
        genetic map build, e.g. hg38, hg19
    out : str
        output prefix

    """
    log_params("haptools-simu-admix", locals())
    assert isinstance(admix_prop, dict)
    admix.tools.haptools_simu_admix(
        pfile=pfile,
        admix_prop=admix_prop,
        pop_col=pop_col,
        mapdir=mapdir,
        n_gen=n_gen,
        n_indiv=n_indiv,
        out_prefix=out,
    )


def download_dependency(
    name: str,
    **kwargs,
):
    """
    Download dependency file or cache data, therefore to avoid downloading on the fly.

    name : str
        name of the dependency. Include the following:
        Software:
            - plink2
            - plink
            - gcta64
            - liftOver
            - hapgen2
            - admix-simu
        Data:
            - genetic_map --build hg38/hg19
            - hapmap3_snps
    kwargs : Dict[str, Any]
        keyword arguments to pass to the download function, e.g. build for genetic_map
    """
    log_params("download-dependency", locals())
    if name in ("plink2", "plink", "gcta64", "liftOver", "hapgen2", "admix-simu"):
        path = admix.tools.get_dependency(name)
        admix.logger.info(f"{name} can be found at {path}")
    elif name == "genetic_map":
        assert "build" in kwargs, "build must be specified for genetic_map"
        admix.tools.get_cache_data(name=name, build=kwargs["build"])
    elif name == "hapmap3_snps":
        admix.tools.get_cache_data(name=name)



import pandas as pd
import numpy as np
import subprocess
from typing import Tuple, List
import os
import glob
from . import get_dependency


def run(cmd: str):
    """Shortcut for running plink commands

    Parameters
    ----------
    cmd : str
        gcta command
    """
    bin_path = get_dependency("gcta64")
    subprocess.check_call(f"{bin_path} {cmd}", shell=True)


def read_grm(file_prefix: str) -> Tuple[np.ndarray, pd.DataFrame, np.ndarray]:
    """read GCTA grm file

    Parameters
    ----------
    file_prefix : str
        prefix of the grm file, <file_prefix>.grm.bin, <file_prefix>.grm.id,
        <file_prefix>.grm.N.bin should be present

    Returns
    -------
    Tuple[np.ndarray, pd.DataFrame, np.ndarray]
        K: grm matrix
        df_id: id data frame
        N: N matrix
    """
    bin_file = file_prefix + ".grm.bin"
    N_file = file_prefix + ".grm.N.bin"
    id_file = file_prefix + ".grm.id"

    df_id = pd.read_csv(id_file, sep="\t", header=None, names=["sample_0", "sample_1"])
    n = df_id.shape[0]
    k = np.asarray(np.fromfile(bin_file, dtype=np.float32), np.float64)
    n_snps = np.asarray(np.fromfile(N_file, dtype=np.float32), np.int64)

    K = np.zeros((n, n))
    K[np.tril_indices_from(K)] = k
    K = K + np.tril(K, -1).T

    return (K, df_id, n_snps)


def write_grm(file_prefix, K, df_id, n_snps):

    bin_file = file_prefix + ".grm.bin"
    N_file = file_prefix + ".grm.N.bin"
    id_file = file_prefix + ".grm.id"

    # id
    df_id.to_csv(id_file, sep="\t", header=None, index=False)
    # bin
    K[np.tril_indices_from(K)].astype(np.float32).tofile(bin_file)
    # N
    n_snps.astype(np.float32).tofile(N_file)


def reml(
    df_pheno: pd.DataFrame,
    out_prefix: str,
    mgrm_path: str = None,
    grm_path: str = None,
    df_covar: pd.DataFrame = None,
    n_thread: int = 4,
    priors: List[float] = None,
    est_fix: bool = False,
    clean_tmp: bool = True,
):
    """Wrapper for GCTA --reml

    Parameters
    ----------
    mgrm_path : str
        mgrm path to file with a list of grms (without .grm.bin, .grm.id, .grm.N.bin)
    grm_path : str
        grm path prefix (without .grm.bin, .grm.id, .grm.N.bin)
    df_pheno : pd.DataFrame
        phenotype data frame
    out_prefix : str
        output prefix
    """
    assert (mgrm_path is None) != (
        grm_path is None
    ), "Either mgrm_path or grm_path must be provided"
    assert df_pheno.shape[1] == 3
    assert df_pheno.columns[0] == "FID"
    assert df_pheno.columns[1] == "IID"

    pheno_path = out_prefix + ".pheno"
    df_pheno.to_csv(pheno_path, index=False, header=False, sep="\t", na_rep="NA")

    cmds = [
        "--reml --reml-no-lrt --reml-no-constrain",
        f"--pheno {pheno_path}",
        f"--out {out_prefix}",
        f"--thread-num {n_thread}",
    ]
    if est_fix:
        cmds[0] += " --reml-est-fix"
    if priors is not None:
        cmds.append("--reml-priors " + " ".join([str(p) for p in priors]))
    if mgrm_path is not None:
        cmds.append(f"--mgrm {mgrm_path}")
    if grm_path is not None:
        cmds.append(f"--grm {grm_path}")

    if df_covar is not None:
        covar_path = out_prefix + ".covar"
        df_covar.to_csv(covar_path, index=False, header=False, sep="\t", na_rep="NA")
        cmds.append(f"--qcovar {covar_path}")

    run(" ".join(cmds))
    if clean_tmp:
        os.remove(pheno_path)
        if df_covar is not None:
            os.remove(covar_path)


def read_reml(path_prefix):
    """
    provide log file of the GCTA output

    read <path_prefix>.log and <path_prefix>.hsq
    """

    # Read log file
    with open(path_prefix + ".log") as f:
        lines = f.readlines()

    ## 1. find "Summary result of REML analysis:"
    line_i = (
        np.where([l.startswith("Summary result of REML analysis:") for l in lines])[
            0
        ].item()
        + 1
    )

    # read all lines before Vp
    est = []
    while 1:
        tmp = lines[line_i].strip().split("\t")
        if tmp[0] == "Vp":
            break
        est.append(tmp)
        line_i += 1
    df_est = pd.DataFrame(est[1:], columns=est[0])
    df_est = df_est.astype({"Variance": float, "SE": float})
    ## 2. find "Sampling variance/covariance of the estimates of variance components:"
    line_i = np.where(
        [
            l.startswith(
                "Sampling variance/covariance of the estimates of variance components:"
            )
            for l in lines
        ]
    )[0].item()
    varcov = [
        lines[l].strip().split("\t")
        for l in range(line_i + 1, line_i + 1 + df_est.shape[0])
    ]
    df_varcov = pd.DataFrame(
        varcov, columns=df_est["Source"].values, index=df_est["Source"].values
    )
    df_varcov = df_varcov.astype(float)

    # Read hsq file
    with open(path_prefix + ".hsq") as f:
        lines = f.readlines()
    # find loglik
    line_i = np.where([l.startswith("logL\t") for l in lines])[0].item()
    loglik = float(lines[line_i].split("\t")[1])

    # find n
    line_i = np.where([l.startswith("n\t") for l in lines])[0].item()
    n_indiv = int(lines[line_i].split("\t")[1])

    # 3. find "Log-likelihood ratio converged."
    return {"est": df_est, "varcov": df_varcov, "loglik": loglik, "n": n_indiv}


def estimate_hsq(dict_est, scale_factor=1):
    """
    Estimate the heritability, and standard error of heritability (using delta method)
    from the results of a single component REML:
    hsq = (var_g * scale_factor) / [var_g * scale_factor + var_e]

    Parameters
    ----------
    dict_est : dict
        dictionary of the results of a single component REML, as returned by read_reml
        {"est": df_est, "varcov": df_varcov}
    scale_factor : float
        scale factor for genotypical variance, default 1

    Returns
    -------
    hsq : float
        heritability
    hsq_var : float
        estimated variance of heritability
    """
    assert len(dict_est["est"]) == 2
    assert np.all(dict_est["est"].Source.values == ["V(G)", "V(e)"])
    est = dict_est["est"].Variance.values
    est_var = dict_est["varcov"].values

    ## method 1
    x, y = est[0], est[1]
    hsq = (x * scale_factor) / (x * scale_factor + y)

    # grad = [y / (x + y)^2, - x / (x + y)^2]
    grad = np.array(
        [
            scale_factor * y / ((scale_factor * x + y) ** 2),
            -scale_factor * x / ((scale_factor * x + y) ** 2),
        ]
    )

    def quad_form(x, A):
        return np.dot(np.dot(x.T, A), x)

    return hsq, quad_form(grad, est_var)


def calculate_hsq_scale(
    weight_file: str, freq_file: str, freq_col: str = "FREQ", index_col: str = "snp"
) -> float:
    """Calculate the heritability scaling factor using weight (prior variance) files
    scale_factor = sum(normalized_weight * 2 * freq * (1 - freq)), where
    normalized_weight = weight / sum(weight).

    Parameters
    ----------
    weight_file : str
        weight file
    freq_file : str
        frequency file, use a wildcard '*' to multiple frequency files
    freq_col : str
        column name of the frequency, default 'FREQ'
    index_col : str
        column name of the index, default 'snp'

    Returns
    -------
    scale_factor : float
        scaling factor

    Notes
    -----
    SNP list in weight_file should be a superset of the SNP list in freq_file.
    """
    df_weight = pd.read_csv(
        weight_file,
        sep="\t",
        index_col=0,
    )

    freq_file_list = glob.glob(freq_file)
    df_freq = (
        pd.concat(
            [
                pd.read_csv(
                    f,
                    delim_whitespace=True,
                )
                for f in freq_file_list
            ]
        ).reset_index(drop=True)
    ).set_index(index_col)

    # df_freq.index should be a superset of df_weight.index
    assert set(df_freq.index).issuperset(set(df_weight.index))
    df_weight["PRIOR_VAR"] /= df_weight["PRIOR_VAR"].sum()
    df_weight[freq_col] = df_freq[freq_col].reindex(df_weight.index)
    df_weight["FREQ_VAR"] = 2 * df_weight[freq_col] * (1 - df_weight[freq_col])

    scale_factor = np.sum(df_weight["PRIOR_VAR"] * df_weight["FREQ_VAR"])

    return scale_factor